# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Inventory Adjustments
"""

from collections import OrderedDict

from wuttaweb.views import MasterView
from wuttaweb.forms import widgets
from wuttaweb.forms.schema import WuttaQuantity

from wuttapos.db.model.products import (
    InventoryAdjustment,
    InventoryAdjustmentType,
)
from wuttapos.server.forms.schema import ProductRef, InventoryAdjustmentTypeRef


def render_grid_product(adjustment, field, value):
    product = adjustment.inventory.product
    return str(product)


class InventoryAdjustmentTypeView(MasterView):
    """
    Master view for Inventory Adjustment Types
    """

    model_class = InventoryAdjustmentType
    model_title = "Inventory Adjustment Type"
    model_title_plural = "Inventory Adjustment Types"

    route_prefix = "inventory_adjustment_types"
    url_prefix = "/inventory/adjustment-types"

    creatable = True
    editable = True
    deletable = True

    grid_columns = [
        "type_code",
        "name",
    ]

    form_fields = [
        "type_code",
        "name",
    ]

    has_rows = True
    row_model_class = InventoryAdjustment
    row_model_title_plural = "Inventory Adjustments"

    row_grid_columns = [
        "product",
        "adjusted",
        "effective_date",
        "adjustment_type",
        "amount",
        "source",
    ]

    rows_sort_defaults = ("adjusted", "desc")

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # links
        g.set_link("type_code")
        g.set_link("name")

    def get_row_grid_data(self, obj):
        adjustment_type = obj
        model = self.app.model
        session = self.app.get_session(adjustment_type)

        return session.query(model.InventoryAdjustment).filter(
            model.InventoryAdjustment.adjustment_type == adjustment_type
        )

    def configure_row_grid(self, grid):
        g = grid
        super().configure_row_grid(g)
        model = self.app.model
        session = self.Session()

        # product
        g.set_renderer("product", render_grid_product)
        g.set_link("product")

        # view action
        def view_url(adjustment, i):
            return self.request.route_url(
                "inventory_adjustments.view", uuid=adjustment.uuid
            )

        g.add_action("view", url=view_url, icon="eye")


class InventoryAdjustmentView(MasterView):
    """
    Master view for Inventory Adjustments
    """

    model_class = InventoryAdjustment
    model_title = "Inventory Adjustment"
    model_title_plural = "Inventory Adjustments"

    route_prefix = "inventory_adjustments"
    url_prefix = "/inventory/adjustments"

    creatable = True
    editable = False
    deletable = False

    grid_columns = [
        "product",
        "adjusted",
        "effective_date",
        "adjustment_type",
        "amount",
        "source",
    ]

    form_fields = [
        "product",
        "effective_date",
        "adjusted",
        "adjustment_type",
        "amount",
        "source",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)
        model = self.app.model
        session = self.Session()

        # product
        g.set_renderer("product", render_grid_product)
        g.set_link("product")

        # adjustment_type
        g.set_joiner(
            "adjustment_type",
            lambda q: q.outerjoin(
                model.InventoryAdjustmentType,
                model.InventoryAdjustmentType.type_code
                == self.model_class.adjustment_type_code,
            ),
        )
        g.set_sorter("adjustment_type", model.InventoryAdjustmentType.name)
        g.remove_filter("adjustment_type_code")
        types = session.query(model.InventoryAdjustmentType).order_by(
            model.InventoryAdjustmentType.name
        )
        choices = OrderedDict([(typ.type_code, typ.name) for typ in types])
        g.set_filter(
            "adjustment_type",
            model.InventoryAdjustmentType.type_code,
            verbs=["equal", "not_equal"],
            choices=choices,
        )

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        model = self.app.model
        session = self.Session()
        adjustment = f.model_instance

        # product
        f.set_node("product", ProductRef(self.request))
        if self.creating:
            if uuid := self.request.GET.get("product"):
                if product := session.get(model.Product, uuid):
                    f.set_default("product", product)
                    f.fields.insert_after("product", "on_hand")
                    f.set_node("on_hand", WuttaQuantity(self.request))
                    f.set_readonly("on_hand")
                    f.set_default(
                        "on_hand",
                        product.inventory.on_hand if product.inventory else None,
                    )
        else:
            f.set_default("product", adjustment.inventory.product)

        # adjustment_type
        f.set_node(
            "adjustment_type",
            InventoryAdjustmentTypeRef(self.request, empty_option=True),
        )
        f.set_required("adjustment_type", False)

        # adjusted
        if self.creating:
            f.remove("adjusted")

        # effective_date
        if self.creating:
            f.remove("effective_date")

        # amount
        f.set_node("amount", WuttaQuantity(self.request))

    def objectify(self, form):
        model = self.app.model
        adjustment = super().objectify(form)

        if self.creating:

            inventory = form.validated["product"].inventory
            if not inventory:
                inventory = model.ProductInventory(product=form.validated["product"])

            adjustment.inventory = inventory
            adjustment.adjusted = self.app.make_utc()
            adjustment.effective_date = self.app.localtime().date()

            if adjustment.adjustment_type == -99:
                adjustment.adjustment_type = None

            inventory.on_hand = (inventory.on_hand or 0) + adjustment.amount

        return adjustment


def defaults(config, **kwargs):
    base = globals()

    InventoryAdjustmentTypeView = kwargs.get(
        "InventoryAdjustmentTypeView", base["InventoryAdjustmentTypeView"]
    )
    InventoryAdjustmentTypeView.defaults(config)

    InventoryAdjustmentView = kwargs.get(
        "InventoryAdjustmentView", base["InventoryAdjustmentView"]
    )
    InventoryAdjustmentView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS config extension
"""

from wuttjamaican.conf import WuttaConfigExtension


class WuttaPosConfigExtension(WuttaConfigExtension):
    """
    The :term`config extension` for WuttaPOS.
    """

    key = "wuttapos"

    def configure(self, config):

        # app info
        config.setdefault(f"{config.appname}.app_title", "WuttaPOS")
        config.setdefault(f"{config.appname}.app_dist", "WuttaPOS")

        # app handler
        config.setdefault(
            f"{config.appname}.app.handler", "wuttapos.app:WuttaPosAppHandler"
        )

        # app model
        config.setdefault(f"{config.appname}.model_spec", "wuttapos.db.model")
        config.setdefault(f"{config.appname}.enum_spec", "wuttapos.enum")

        # # auth handler
        # config.setdefault(
        #     f"{config.appname}.auth.handler", "wuttapos.auth:WuttaPosAuthHandler"
        # )

        # server menu handler
        config.setdefault(
            f"{config.appname}.web.menus.handler.spec",
            "wuttapos.server.menus:WuttaPosMenuHandler",
        )

        # # web app libcache
        # #config.setdefault('wuttaweb.static_libcache.module', 'wuttapos.server.static')



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Clientele Handler
"""

# from collections import OrderedDict
import logging
import uuid as _uuid

# import warnings

from sqlalchemy import orm

from wuttjamaican.app import GenericHandler


log = logging.getLogger(__name__)


class ClienteleHandler(GenericHandler):
    """
    Base class and default implementation for clientele handlers.
    """

    # def get_customer(self, obj):
    #     """
    #     Return the Customer associated with the given object, if any.
    #     """
    #     model = self.model

    #     if isinstance(obj, model.Customer):
    #         return obj

    #     else:
    #         person = self.app.get_person(obj)
    #         if person:
    #             # TODO: all 3 options below are indeterminate, since it's
    #             # *possible* for a person to hold multiple accounts
    #             # etc. but not sure how to fix in a generic way?  maybe
    #             # just everyone must override as needed
    #             if person.customer_accounts:
    #                 return person.customer_accounts[0]
    #             for shopper in person.customer_shoppers:
    #                 if shopper.shopper_number == 1:
    #                     return shopper.customer
    #             # legacy fallback
    #             if person.customers:
    #                 return person.customers[0]

    # def make_customer(self, person, **kwargs):
    #     """
    #     Create and return a new customer record.
    #     """
    #     session = self.app.get_session(person)
    #     customer = self.model.Customer()
    #     customer.name = person.display_name
    #     customer.account_holder = person
    #     session.add(customer)
    #     session.flush()
    #     session.refresh(person)
    #     return customer

    def locate_customer_for_entry(self, session, entry, **kwargs):
        """
        This method aims to provide sane default logic for locating a
        :class:`~rattail.db.model.customers.Customer` record for the
        given "entry" value.

        The default logic here will try to honor the "configured"
        customer field, and prefer that when attempting the lookup.

        :param session: Reference to current DB session.

        :param entry: Value to use for lookup.  This is most often a
           simple string, but the method can handle a few others.  For
           instance it is common to read values from a spreadsheet,
           and sometimes those come through as integers etc.

        :param lookup_fields: Optional list of fields to use for
           lookup.  The default value is ``['uuid', '_customer_key_']``
           which means to lookup by UUID as well as "customer key"
           field, which is configurable.  You can include any of the
           following in ``lookup_fields``:

           * ``uuid``
           * ``_customer_key_`` - :meth:`locate_customer_for_key`

        :returns: First :class:`~rattail.db.model.customers.Customer`
           instance found if there was a match; otherwise ``None``.
        """
        model = self.app.model
        if not entry:
            return None

        # figure out which fields we should match on
        # TODO: let config declare default lookup_fields
        lookup_fields = kwargs.get(
            "lookup_fields",
            [
                "uuid",
                "customer_id",
            ],
        )

        # try to locate customer by uuid before other, more specific key
        if "uuid" in lookup_fields:
            if isinstance(entry, (_uuid.UUID, str)):
                customer = session.get(model.Customer, entry)
                if customer:
                    return customer

        lookups = {
            "uuid": None,
            "customer_id": self.locate_customer_for_id,
        }

        for field in lookup_fields:
            if field in lookups:
                lookup = lookups[field]
                if lookup:
                    customer = lookup(session, entry, **kwargs)
                    if customer:
                        return customer
            else:
                log.warning("unknown lookup field: %s", field)

    def locate_customer_for_id(self, session, entry, **kwargs):
        """
        Locate the customer which matches the given ID.

        This will do a lookup on the
        :attr:`rattail.db.model.customers.Customer.id` field only.

        Note that instead of calling this method directly, you might
        consider calling :meth:`locate_customer_for_key()` instead.

        :param session: Current session for Rattail DB.

        :param entry: Customer ID value as string.

        :returns: First :class:`~rattail.db.model.customers.Customer`
           instance found if there was a match; otherwise ``None``.
        """
        if not entry:
            return None

        # assume entry is string
        entry = str(entry)

        model = self.app.model
        try:
            return (
                session.query(model.Customer)
                .filter(model.Customer.customer_id == entry)
                .one()
            )
        except orm.exc.NoResultFound:
            return None

    def search_customers(self, session, entry, **kwargs):
        """
        Perform a customer search across multiple fields, and return
        results as JSON data rows.
        """
        model = self.app.model
        final_results = []

        # first we'll attempt "lookup" logic..

        lookup_fields = kwargs.get(
            "lookup_fields",
            ["customer_id"],
        )

        if lookup_fields:
            customer = self.locate_customer_for_entry(
                session, entry, lookup_fields=lookup_fields
            )
            if customer:
                final_results.append(customer)

        # then we'll attempt "search" logic..

        search_fields = kwargs.get(
            "search_fields",
            [
                "name",
                "email_address",
                "phone_number",
            ],
        )

        searches = {
            "name": self.search_customers_for_name,
            "email_address": self.search_customers_for_email_address,
            "phone_number": self.search_customers_for_phone_number,
        }

        for field in search_fields:
            if field in searches:
                search = searches[field]
                if search:
                    customers = search(session, entry, **kwargs)
                    final_results.extend(customers)
            else:
                log.warning("unknown search field: %s", field)

        return [self.normalize_customer(c) for c in final_results]

    def search_customers_for_name(self, session, entry, **kwargs):
        model = self.app.model
        entry = entry.lower()

        return (
            session.query(model.Customer)
            .filter(model.Customer.name.ilike(f"%{entry}%"))
            .all()
        )

    def search_customers_for_email_address(self, session, entry, **kwargs):
        model = self.app.model
        entry = entry.lower()

        return (
            session.query(model.Customer)
            .filter(model.Customer.email_address.ilike(f"%{entry}%"))
            .all()
        )

    def search_customers_for_phone_number(self, session, entry, **kwargs):
        model = self.app.model
        entry = entry.lower()

        return (
            session.query(model.Customer)
            .filter(model.Customer.phone_number.ilike(f"%{entry}%"))
            .all()
        )

    def normalize_customer(self, customer):
        """
        Normalize the given customer to a JSON-serializable dict.
        """
        return {
            "uuid": customer.uuid,
            "customer_id": customer.customer_id,
            "name": customer.name,
            "phone_number": customer.phone_number,
            "email_address": customer.email_address,
            "_str": str(customer),
        }

    def get_customer_info_markdown(self, customer, **kwargs):
        """
        Returns a Markdown string containing pertinent info about a
        given customer account.
        """
        return (
            f"Customer ID: {customer.customer_id}\n\n"
            f"Name: {customer.name}\n\n"
            f"Phone: {customer.phone_number or ''}\n\n"
            f"Email: {customer.email_address or ''}\n\n"
        )



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - login view
"""

import flet as ft

from .base import WuttaView
from wuttapos.terminal.controls.loginform import WuttaLoginForm


class LoginView(WuttaView):
    """
    Main POS view for WuttaPOS
    """

    def build_controls(self):
        title = self.app.get_title()

        controls = [
            ft.Row(
                [self.make_logo_image(height=200)],
                alignment=ft.MainAxisAlignment.CENTER,
            ),
            ft.Row(
                [
                    ft.Text(
                        value=f"Welcome to {title}", weight=ft.FontWeight.BOLD, size=28
                    )
                ],
                alignment=ft.MainAxisAlignment.CENTER,
            ),
            ft.Row(),
            ft.Row(),
            ft.Row(),
            WuttaLoginForm(
                self.config,
                on_login_failure=self.login_failure,
                on_authz_failure=self.authz_failure,
                on_login_success=self.login_success,
            ),
        ]

        return [
            self.build_header(),
            ft.Column(
                controls=controls, expand=True, alignment=ft.MainAxisAlignment.CENTER
            ),
        ]

    def login_failure(self, e):
        self.show_snackbar("Login failed!", bgcolor="yellow")
        self.page.update()

    def authz_failure(self, user, user_display):
        self.show_snackbar(
            f"User not allowed to ring sales: {user_display}", bgcolor="yellow"
        )
        self.page.update()

    def login_success(self, user, user_display):
        self.page.session.set("user_uuid", user.uuid.hex)
        self.page.session.set("user_display", user_display)

        # TODO: hacky but works for now
        if not self.config.production():
            self.page.client_storage.set("user_uuid", user.uuid.hex)

        self.page.go("/pos")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - POS view
"""

import decimal
import logging
import time

import flet as ft

from .base import WuttaView
from wuttapos.terminal.controls.loginform import WuttaLoginForm
from wuttapos.terminal.controls.custlookup import WuttaCustomerLookup
from wuttapos.terminal.controls.itemlookup import WuttaProductLookup
from wuttapos.terminal.controls.itemlookup_dept import WuttaProductLookupByDepartment
from wuttapos.terminal.controls.deptlookup import WuttaDepartmentLookup
from wuttapos.terminal.controls.txnlookup import WuttaTransactionLookup
from wuttapos.terminal.controls.txnitem import WuttaTxnItem
from wuttapos.terminal.controls.menus.tenkey import WuttaTenkeyMenu


log = logging.getLogger(__name__)


class POSView(WuttaView):
    """
    Main POS view for WuttaPOS
    """

    # TODO: should be configurable?
    default_button_size = 100
    default_font_size = 40

    disabled_bgcolor = "#aaaaaa"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # keep a list of "informed" controls - i.e. child controls
        # within this view, which need to stay abreast of global
        # changes to the transaction, customer etc.
        self.informed_controls = []
        if hasattr(self, "header"):
            self.informed_controls.append(self.header)

    def informed_refresh(self):
        for control in self.informed_controls:
            control.informed_refresh()

    def reset(self, e=None, clear_quantity=True):
        """
        This is a convenience method, meant only to clear the main
        input and set focus to it.  Will also update() the page.

        The ``e`` arg is ignored and accepted only so this method may
        be registered as an event handler, e.g. ``on_cancel``.
        """
        # clear set (@) quantity
        if clear_quantity:
            if self.set_quantity.data:
                self.set_quantity.data = None
                self.set_quantity.value = None
                self.set_quantity_button.visible = True

        # clear/focus main input
        self.main_input.value = ""
        self.main_input.focus()

        self.page.update()

    def set_customer(self, customer, batch=None, user=None):
        session = self.app.get_session(customer)
        if not batch:
            batch = self.get_current_batch(session)
        if user:
            user = session.get(user.__class__, user.uuid)
        else:
            user = self.get_current_user(session)

        handler = self.get_batch_handler()
        handler.set_customer(batch, customer, user=user)

        self.page.session.set("txn_display", handler.get_screen_txn_display(batch))
        self.page.session.set("cust_uuid", customer.uuid)
        self.page.session.set(
            "cust_display", handler.get_screen_cust_display(customer=customer)
        )
        self.informed_refresh()
        self.refresh_totals(batch)

        self.show_snackbar(f"CUSTOMER SET: {customer}", bgcolor="green")

    def refresh_totals(self, batch):
        reg = ft.TextStyle(size=22)
        bold = ft.TextStyle(size=24, weight=ft.FontWeight.BOLD)

        self.subtotals.spans.clear()

        sales_total = batch.sales_total or 0
        self.subtotals.spans.append(ft.TextSpan("Sales  ", style=reg))
        total = self.app.render_currency(sales_total)
        self.subtotals.spans.append(ft.TextSpan(total, style=bold))

        tax_total = 0
        # for tax_id, tax in sorted(txn["taxes"].items()):
        #     if tax["tax_total"]:
        #         self.subtotals.spans.append(
        #             ft.TextSpan(f"    Tax {tax_id}  ", style=reg)
        #         )
        #         total = self.app.render_currency(tax["tax_total"])
        #         self.subtotals.spans.append(ft.TextSpan(total, style=bold))
        #         tax_total += tax["tax_total"]

        tender_total = 0
        # tender_total = sum(
        #     [tender["tender_total"] for tender in txn["tenders"].values()]
        # )
        if tender_total:
            self.subtotals.spans.append(ft.TextSpan(f"    Tend  ", style=reg))
            total = self.app.render_currency(tender_total)
            self.subtotals.spans.append(ft.TextSpan(total, style=bold))

        self.fs_balance.spans.clear()
        # fs_total = txn["foodstamp"]
        fs_total = 0
        fs_balance = fs_total + tender_total
        if fs_balance:
            self.fs_balance.spans.append(ft.TextSpan("FS  ", style=reg))
            total = self.app.render_currency(fs_balance)
            self.fs_balance.spans.append(ft.TextSpan(total, style=bold))

        self.balances.spans.clear()
        total_due = sales_total + tax_total + tender_total
        total_due = self.app.render_currency(total_due)
        self.balances.spans.append(ft.TextSpan("    ", style=reg))
        self.balances.spans.append(
            ft.TextSpan(
                total_due, style=ft.TextStyle(size=40, weight=ft.FontWeight.BOLD)
            )
        )

        self.totals_row.bgcolor = "orange"

    def attempt_add_product(self, uuid=None, record_badscan=False):
        model = self.app.model
        enum = self.app.enum
        session = self.app.make_session()
        handler = self.get_batch_handler()
        user = self.get_current_user(session)
        batch = self.get_current_batch(session, user=user)
        entry = self.main_input.value

        quantity = 1
        if self.set_quantity.data is not None:
            quantity = self.set_quantity.data

        product = None
        item_entry = entry
        if uuid:
            product = session.get(model.Product, uuid)
            assert product
            item_entry = product.product_id or uuid

        try:
            row = handler.process_entry(
                batch,
                product or entry,
                quantity=quantity,
                item_entry=item_entry,
                user=user,
            )
        except Exception as error:
            session.rollback()
            self.show_snackbar(f"ERROR: {error}", bgcolor="yellow")
            row = None

        else:

            if row:
                session.commit()

                if row.row_type == enum.POS_ROW_TYPE_BADPRICE:
                    self.show_snackbar(
                        f"Product has invalid price: {row.item_entry}", bgcolor="yellow"
                    )

                else:
                    session.expunge(row)
                    self.add_row_item(row, scroll=True)
                    self.refresh_totals(batch)
                    self.reset()

            else:

                if record_badscan:
                    handler.record_badscan(batch, entry, quantity=quantity, user=user)

                self.show_snackbar(f"PRODUCT NOT FOUND: {entry}", bgcolor="yellow")

            session.commit()
            self.refresh_totals(batch)

        session.close()
        self.page.update()
        return bool(row)

    def item_lookup(self, value=None):

        def select(uuid):
            self.attempt_add_product(uuid=uuid)
            dlg.open = False
            self.reset()

        def cancel(e):
            dlg.open = False
            self.reset(clear_quantity=False)

        dlg = ft.AlertDialog(
            modal=True,
            title=ft.Text("Product Lookup"),
            content=WuttaProductLookup(
                self.config, initial_search=value, on_select=select, on_cancel=cancel
            ),
        )

        # self.page.open(dlg)

        self.page.dialog = dlg
        dlg.open = True
        self.page.update()

    def customer_lookup(self, value=None, user=None):
        model = self.app.model

        def select(uuid):
            session = self.app.make_session()
            customer = session.get(model.Customer, uuid)
            self.set_customer(customer, user=user)
            session.commit()
            session.close()

            dlg.open = False
            self.reset()

        def cancel(e):
            dlg.open = False
            self.reset()

        dlg = ft.AlertDialog(
            modal=True,
            title=ft.Text("Customer Lookup"),
            content=WuttaCustomerLookup(
                self.config, initial_search=value, on_select=select, on_cancel=cancel
            ),
        )

        # self.page.open(dlg)

        self.page.dialog = dlg
        dlg.open = True
        self.page.update()

    def customer_info(self):
        model = self.app.model
        session = self.app.make_session()
        clientele = self.app.get_clientele_handler()

        entry = self.main_input.value
        if entry:
            different = True
            customer = clientele.locate_customer_for_entry(session, entry)
            if not customer:
                session.close()
                self.show_snackbar(f"CUSTOMER NOT FOUND: {entry}", bgcolor="yellow")
                self.page.update()
                return

        else:
            different = False
            customer = session.get(model.Customer, self.page.session.get("cust_uuid"))
            assert customer

        info = clientele.get_customer_info_markdown(customer)
        session.close()

        def close(e):
            dlg.open = False
            self.reset()

        font_size = self.default_font_size * 0.8
        dlg = ft.AlertDialog(
            # modal=True,
            title=ft.Text("Customer Info"),
            content=ft.Container(
                ft.Column(
                    [
                        ft.Container(
                            content=ft.Text(
                                "NOTE: this is a DIFFERENT customer than the txn has!"
                            ),
                            bgcolor="yellow",
                            visible=different,
                        ),
                        ft.Divider(),
                        ft.Container(
                            theme_mode=ft.ThemeMode.SYSTEM,
                            theme=ft.Theme(
                                text_theme=ft.TextTheme(
                                    body_medium=ft.TextStyle(
                                        size=24,
                                        color="black",
                                    )
                                )
                            ),
                            content=ft.Markdown(info),
                        ),
                    ],
                    height=500,
                    width=500,
                )
            ),
            actions=[
                ft.Container(
                    content=ft.Text("Close", size=font_size, weight=ft.FontWeight.BOLD),
                    height=self.default_button_size * 0.8,
                    width=self.default_button_size * 1.2,
                    alignment=ft.alignment.center,
                    bgcolor="blue",
                    border=ft.border.all(1, "black"),
                    border_radius=ft.border_radius.all(5),
                    on_click=close,
                ),
            ],
            # actions_alignment=ft.MainAxisAlignment.END,
        )

        # self.page.open(dlg)

        self.page.dialog = dlg
        dlg.open = True
        self.page.update()

    def customer_prompt(self):

        def view_info(e):
            dlg.open = False
            self.page.update()

            # cf. https://github.com/flet-dev/flet/issues/1670
            time.sleep(0.1)

            self.customer_info()

        def remove(e):
            dlg.open = False
            self.page.update()

            # cf. https://github.com/flet-dev/flet/issues/1670
            time.sleep(0.1)

            self.remove_customer_prompt()

        def replace(e):
            dlg.open = False
            self.page.update()

            # nb. do this just in case we must show login dialog
            # cf. https://github.com/flet-dev/flet/issues/1670
            time.sleep(0.1)

            self.authorized_action(
                "pos.swap_customer", self.replace_customer, message="Replace Customer"
            )

        def cancel(e):
            dlg.open = False
            self.reset()

        font_size = self.default_font_size * 0.8
        dlg = ft.AlertDialog(
            # modal=True,
            title=ft.Text("Customer Already Selected"),
            content=ft.Text("What would you like to do?", size=20),
            actions=[
                ft.Container(
                    content=ft.Text(
                        "Remove", size=font_size, weight=ft.FontWeight.BOLD
                    ),
                    height=self.default_button_size,
                    width=self.default_button_size * 2.5,
                    alignment=ft.alignment.center,
                    bgcolor="red",
                    border=ft.border.all(1, "black"),
                    border_radius=ft.border_radius.all(5),
                    on_click=remove,
                ),
                ft.Container(
                    content=ft.Text(
                        "Replace",
                        size=font_size,
                        color="black",
                        weight=ft.FontWeight.BOLD,
                    ),
                    height=self.default_button_size,
                    width=self.default_button_size * 2.5,
                    alignment=ft.alignment.center,
                    bgcolor="yellow",
                    border=ft.border.all(1, "black"),
                    border_radius=ft.border_radius.all(5),
                    on_click=replace,
                ),
                ft.Container(
                    content=ft.Text(
                        "View Info", size=font_size, weight=ft.FontWeight.BOLD
                    ),
                    height=self.default_button_size,
                    width=self.default_button_size * 2.5,
                    alignment=ft.alignment.center,
                    bgcolor="blue",
                    border=ft.border.all(1, "black"),
                    border_radius=ft.border_radius.all(5),
                    on_click=view_info,
                ),
                ft.Container(
                    content=ft.Text(
                        "Cancel",
                        size=font_size,
                        # color='black',
                        weight=ft.FontWeight.BOLD,
                    ),
                    height=self.default_button_size,
                    width=self.default_button_size * 2.5,
                    alignment=ft.alignment.center,
                    border=ft.border.all(1, "black"),
                    border_radius=ft.border_radius.all(5),
                    on_click=cancel,
                ),
            ],
        )

        # self.page.open(dlg)

        self.page.dialog = dlg
        dlg.open = True
        self.page.update()

    def remove_customer_prompt(self):

        def remove(e):
            dlg.open = False
            self.page.update()

            # nb. do this just in case we must show login dialog
            # cf. https://github.com/flet-dev/flet/issues/1670
            time.sleep(0.1)

            self.authorized_action(
                "pos.del_customer", self.remove_customer, message="Remove Customer"
            )

        def cancel(e):
            dlg.open = False
            self.reset()

        font_size = self.default_font_size * 0.8
        dlg = ft.AlertDialog(
            title=ft.Text("Remove Customer"),
            content=ft.Text(
                "Really remove the customer from this transaction?", size=20
            ),
            actions=[
                ft.Container(
                    content=ft.Text(
                        "Yes, Remove", size=font_size, weight=ft.FontWeight.BOLD
                    ),
                    height=self.default_button_size,
                    width=self.default_button_size * 2.5,
                    alignment=ft.alignment.center,
                    bgcolor="red",
                    border=ft.border.all(1, "black"),
                    border_radius=ft.border_radius.all(5),
                    on_click=remove,
                ),
                ft.Container(
                    content=ft.Text(
                        "Cancel", size=font_size, weight=ft.FontWeight.BOLD
                    ),
                    height=self.default_button_size,
                    width=self.default_button_size * 2.5,
                    alignment=ft.alignment.center,
                    border=ft.border.all(1, "black"),
                    border_radius=ft.border_radius.all(5),
                    on_click=cancel,
                ),
            ],
        )

        # self.page.open(dlg)

        self.page.dialog = dlg
        dlg.open = True
        self.page.update()

    def remove_customer(self, user):
        session = self.app.make_session()
        handler = self.get_batch_handler()
        batch = self.get_current_batch(session)
        user = session.get(user.__class__, user.uuid)
        handler.set_customer(batch, None, user=user)
        session.commit()
        session.close()

        self.page.session.set("cust_uuid", None)
        self.page.session.set("cust_display", None)
        self.informed_refresh()
        self.show_snackbar("CUSTOMER REMOVED", bgcolor="yellow")
        self.reset()

    def replace_customer(self, user):
        entry = self.main_input.value
        if entry:
            if not self.attempt_set_customer(entry, user=user):
                self.customer_lookup(entry, user=user)
        else:
            self.customer_lookup(user=user)

    def attempt_set_customer(self, entry=None, user=None):
        session = self.app.make_session()

        customer = self.app.get_clientele_handler().locate_customer_for_entry(
            session, entry
        )
        if customer:

            self.set_customer(customer, user=user)
            self.reset()

        else:  # customer not found
            self.show_snackbar(f"CUSTOMER NOT FOUND: {entry}", bgcolor="yellow")
            # TODO: should use reset() here?
            self.main_input.focus()
            self.page.update()

        session.commit()
        session.close()
        return bool(customer)

    def build_controls(self):

        # handler = self.get_transaction_handler()
        # corepos = self.app.get_corepos_handler()
        # op_session = corepos.make_session_lane_op()
        # self.tender_cash = handler.get_tender(op_session, 'cash')
        # self.tender_check = handler.get_tender(op_session, 'check')
        # self.tender_foodstamp = handler.get_tender(op_session, 'foodstamp')
        # op_session.expunge_all()
        # op_session.close()

        self.main_input = ft.TextField(
            on_submit=self.main_submit,
            text_size=24,
            text_style=ft.TextStyle(weight=ft.FontWeight.BOLD),
            autofocus=True,
        )

        self.selected_item = None
        self.items = ft.ListView(
            item_extent=50,
            height=800,
        )

        self.subtotals = ft.Text(spans=[])
        self.fs_balance = ft.Text(spans=[])
        self.balances = ft.Text(spans=[])

        self.totals_row = ft.Container(
            ft.Row(
                [
                    self.subtotals,
                    ft.Row(
                        [
                            self.fs_balance,
                            self.balances,
                        ],
                    ),
                ],
                alignment=ft.MainAxisAlignment.SPACE_BETWEEN,
            ),
            padding=ft.padding.only(10, 0, 10, 0),
        )

        self.items_column = ft.Column(
            controls=[
                ft.Container(content=self.items, padding=ft.padding.only(10, 0, 10, 0)),
                self.totals_row,
            ],
            expand=1,
        )

        def backspace_click(e):
            if self.main_input.value:
                self.main_input.value = self.main_input.value[:-1]
            self.main_input.focus()
            self.page.update()

        def clear_entry_click(e):
            if self.main_input.value:
                self.main_input.value = ""
            elif self.set_quantity.data is not None:
                self.set_quantity.data = None
                self.set_quantity.value = None
                self.set_quantity_button.visible = True
            elif self.selected_item:
                self.clear_item_selection()
            self.main_input.focus()
            self.page.update()

        self.set_quantity = ft.Text(
            value=None, data=None, weight=ft.FontWeight.BOLD, size=40
        )

        self.set_quantity_button = self.make_button(
            "@",
            font_size=40,
            height=70,
            width=70,
            bgcolor="green",
            on_click=self.set_quantity_click,
        )

        spec = self.config.get(
            "wuttapos.menus.master.spec",
            default="wuttapos.terminal.controls.menus.master:WuttaMenuMaster",
        )
        factory = self.app.load_object(spec)
        self.menu_master = factory(self.config, pos=self)

        return [
            self.build_header(),
            ft.Row(
                [
                    self.make_logo_image(height=80),
                    ft.Row(
                        [
                            ft.Row(
                                [
                                    self.set_quantity,
                                    self.set_quantity_button,
                                ],
                            ),
                            self.main_input,
                            self.make_button(
                                "⌫",
                                font_size=40,
                                bgcolor="green",
                                height=70,
                                width=70,
                                on_click=backspace_click,
                            ),
                            self.make_button(
                                "CE",
                                font_size=40,
                                bgcolor="green",
                                height=70,
                                width=70,
                                on_click=clear_entry_click,
                            ),
                        ],
                        alignment=ft.MainAxisAlignment.CENTER,
                        expand=True,
                    ),
                ],
            ),
            ft.Row(),
            ft.Row(),
            ft.Row(),
            ft.Row(
                [
                    self.items_column,
                    self.menu_master,
                ],
                vertical_alignment=ft.CrossAxisAlignment.START,
            ),
        ]

    def make_button(self, *args, **kwargs):
        kwargs.setdefault("pos", self)
        return super().make_button(*args, **kwargs)

    def make_text(self, *args, **kwargs):
        kwargs.setdefault("weight", ft.FontWeight.BOLD)
        kwargs.setdefault("size", 24)
        return ft.Text(*args, **kwargs)

    def set_quantity_click(self, e):
        quantity = self.main_input.value
        valid = False

        if self.set_quantity.data is not None:
            quantity = self.set_quantity.data
            self.show_snackbar(f"QUANTITY ALREADY SET: {quantity}", bgcolor="yellow")

        else:
            try:
                quantity = decimal.Decimal(quantity)
                valid = True
            except decimal.InvalidOperation:
                pass

            if valid and quantity:
                self.set_quantity.data = quantity
                self.set_quantity.value = self.app.render_quantity(quantity) + " @ "
                self.set_quantity_button.visible = False
                self.main_input.value = ""
                self.main_input.focus()

            else:
                self.show_snackbar(f"INVALID @ QUANTITY: {quantity}", bgcolor="yellow")

        self.page.update()

    def suspend_transaction(self, user):
        session = self.app.make_session()
        batch = self.get_current_batch(session)
        user = session.get(user.__class__, user.uuid)
        handler = self.get_batch_handler()

        handler.suspend_transaction(batch, user)

        session.commit()
        session.close()
        self.clear_all()
        self.reset()

    def get_current_user(self, session):
        model = self.app.model
        uuid = self.page.session.get("user_uuid")
        if uuid:
            return session.get(model.User, uuid)

    def refresh_training(self):
        if self.page.session.get("training"):
            self.bgcolor = "#E4D97C"
        else:
            self.bgcolor = None

    def get_current_batch(self, session, user=None, create=True):
        handler = self.get_batch_handler()

        if not user:
            user = self.get_current_user(session)

        training = bool(self.page.session.get("training"))
        batch, created = handler.get_current_batch(
            user, training_mode=training, create=create, return_created=True
        )

        if created:
            self.page.session.set("txn_display", handler.get_screen_txn_display(batch))
            self.informed_refresh()

        return batch

    def did_mount(self):
        session = self.app.make_session()

        if batch := self.get_current_batch(session, create=False):
            self.load_batch(batch)
        else:
            self.page.session.set("txn_display", None)
            self.page.session.set("cust_uuid", None)
            self.page.session.set("cust_display", None)
            self.informed_refresh()

        self.refresh_training()

        # TODO: i think commit() was for when it auto-created the
        # batch, so that can go away now..right?
        # session.commit()
        session.close()
        self.page.update()

    def load_batch(self, batch):
        """
        Load the given data as the current transaction.
        """
        session = self.app.get_session(batch)
        handler = self.get_batch_handler()
        self.page.session.set("txn_display", handler.get_screen_txn_display(batch))
        self.page.session.set(
            "cust_uuid", batch.customer.uuid if batch.customer else None
        )
        self.page.session.set(
            "cust_display", handler.get_screen_cust_display(batch=batch)
        )

        self.items.controls.clear()
        for row in batch.rows:
            session.expunge(row)
            self.add_row_item(row)
        self.items.scroll_to(offset=-1, duration=100)

        self.refresh_totals(batch)
        self.informed_refresh()

    def not_supported(self, e=None, feature=None):

        # test error handler
        if e.control.data and e.control.data.get("error"):
            raise RuntimeError("NOT YET SUPPORTED")

        text = "NOT YET SUPPORTED"
        if not feature and e:
            feature = e.control.content.value.replace("\n", " ")
        if feature:
            text += f": {feature}"
        self.show_snackbar(text, bgcolor="yellow")
        self.page.update()

    def require_decimal(self, value):
        try:
            amount = decimal.Decimal(value)
        except decimal.InvalidOperation:
            self.show_snackbar(f"Amount is not valid: {value}", bgcolor="yellow")
            return False

        if "." not in value:
            self.show_snackbar(f"Decimal point required: {value}", bgcolor="yellow")
            return False

        return amount

    def adjust_price(self, user):
        enum = self.app.enum

        def cancel(e):
            dlg.open = False
            self.main_input.focus()
            self.page.update()

        def clear(e):
            price_override.value = ""
            price_override.focus()
            self.page.update()

        def tenkey_char(key):
            price_override.value = f"{price_override.value or ''}{key}"
            self.page.update()

        def confirm(e):
            price = self.require_decimal(price_override.value)
            if price is False:
                self.main_input.focus()
                self.page.update()
                return

            dlg.open = False

            session = self.app.make_session()
            user = self.get_current_user(session)
            handler = self.get_batch_handler()

            row = self.selected_item.data["row"]
            row = session.get(row.__class__, row.uuid)

            new_row = handler.override_price(row, user, price)
            session.commit()

            # update screen to reflect new balance
            batch = row.batch
            self.refresh_totals(batch)

            # update item display
            session.expunge(row)
            self.selected_item.data["row"] = row
            self.selected_item.content.row = row
            self.selected_item.content.refresh()
            self.items.update()

            session.close()
            self.clear_item_selection()
            self.reset()

        row = self.selected_item.data["row"]

        price = f"{row.txn_price:0.2f}"
        if self.main_input.value:
            try:
                price = decimal.Decimal(self.main_input.value)
            except decimal.InvalidOperation:
                pass
            else:
                price = f"{price:0.2f}"

        price_override = ft.TextField(
            value=price,
            text_size=32,
            text_style=ft.TextStyle(weight=ft.FontWeight.BOLD),
            autofocus=True,
            on_submit=confirm,
        )

        current_price = self.app.render_currency(row.cur_price)
        if current_price:
            current_price += " [{}]".format(
                enum.PRICE_TYPE.get(row.cur_price_type, row.cur_price_type)
            )

        dlg = ft.AlertDialog(
            modal=True,
            title=ft.Text("Adjust Price"),
            content=ft.Container(
                ft.Column(
                    [
                        ft.Divider(),
                        ft.Row(
                            [
                                ft.Text(
                                    "Reg Price:", size=32, weight=ft.FontWeight.BOLD
                                ),
                                ft.Text(
                                    self.app.render_currency(row.reg_price),
                                    size=32,
                                    weight=ft.FontWeight.BOLD,
                                ),
                            ],
                        ),
                        ft.Row(),
                        ft.Row(
                            [
                                ft.Text(
                                    "Cur Price:", size=32, weight=ft.FontWeight.BOLD
                                ),
                                ft.Text(
                                    current_price, size=32, weight=ft.FontWeight.BOLD
                                ),
                            ],
                        ),
                        ft.Row(),
                        ft.Row(),
                        ft.Row(
                            [
                                ft.Text(
                                    "Txn Price:", size=32, weight=ft.FontWeight.BOLD
                                ),
                                ft.VerticalDivider(),
                                ft.Text("$", size=32, weight=ft.FontWeight.BOLD),
                                price_override,
                            ],
                        ),
                        ft.Row(),
                        ft.Row(),
                        ft.Row(
                            [
                                WuttaTenkeyMenu(
                                    self.config,
                                    simple=True,
                                    on_char=tenkey_char,
                                    on_enter=confirm,
                                ),
                                self.make_button(
                                    "Clear",
                                    height=self.default_button_size * 0.8,
                                    width=self.default_button_size * 1.2,
                                    on_click=clear,
                                ),
                            ],
                            alignment=ft.MainAxisAlignment.SPACE_BETWEEN,
                            vertical_alignment=ft.CrossAxisAlignment.START,
                        ),
                    ],
                ),
                height=700,
                width=550,
            ),
            actions=[
                self.make_button(
                    "Cancel",
                    height=self.default_button_size * 0.8,
                    width=self.default_button_size * 1.2,
                    on_click=cancel,
                ),
                self.make_button(
                    "Confirm",
                    bgcolor="blue",
                    height=self.default_button_size * 0.8,
                    width=self.default_button_size * 1.2,
                    on_click=confirm,
                ),
            ],
            actions_alignment=ft.MainAxisAlignment.END,
        )

        # self.page.open(dlg)

        self.page.dialog = dlg
        dlg.open = True
        self.page.update()

    def toggle_training_mode(self, user):
        was_training = self.page.session.get("training")
        now_training = not was_training

        # TODO: hacky but works for now
        if not self.config.production():
            self.page.client_storage.set("training", now_training)

        self.page.session.set("training", now_training)

        self.refresh_training()
        self.informed_refresh()
        self.reset()

    def kick_drawer(self):
        self.show_snackbar("TODO: Drawer Kick", bgcolor="yellow")
        self.page.update()

    def add_row_item(self, row, scroll=False):
        enum = self.app.enum

        # TODO: row types ugh
        if row.row_type not in (
            enum.POS_ROW_TYPE_SELL,
            enum.POS_ROW_TYPE_OPEN_RING,
            enum.POS_ROW_TYPE_TENDER,
            enum.POS_ROW_TYPE_CHANGE_BACK,
        ):
            return

        self.items.controls.append(
            ft.Container(
                content=WuttaTxnItem(self.config, row),
                border=ft.border.only(bottom=ft.border.BorderSide(1, "gray")),
                padding=ft.padding.only(5, 5, 5, 5),
                on_click=self.list_item_click,
                data={"row": row},
                key=row.uuid,
                bgcolor="white",
            )
        )

        if scroll:
            self.items.scroll_to(offset=-1, duration=100)

    def list_item_click(self, e):
        self.select_txn_item(e.control)

    def select_txn_item(self, item):
        if self.selected_item:
            self.clear_item_selection()

        self.selected_item = item
        self.selected_item.bgcolor = "blue"
        self.page.update()

    def authorized_action(self, perm, action, cancel=None, message=None):
        auth = self.app.get_auth_handler()

        # current user is assumed if they have the perm
        session = self.app.make_session()
        user = self.get_current_user(session)
        has_perm = auth.has_permission(session, user, perm)
        session.expunge(user)
        session.close()
        if has_perm:
            action(user)
            return

        # otherwise must prompt for different user credentials...

        def login_cancel(e):
            dlg.open = False
            if cancel:
                cancel()
            self.reset()

        def login_failure(e):
            self.show_snackbar("Login failed", bgcolor="yellow")
            self.page.update()

        def authz_failure(user, user_display):
            self.show_snackbar(
                f"User does not have permission: {user_display}", bgcolor="yellow"
            )
            self.page.update()

        def login_success(user, user_display):
            dlg.open = False
            self.page.update()

            # nb. just in case next step requires a dialog
            # cf. https://github.com/flet-dev/flet/issues/1670
            time.sleep(0.1)

            action(user)
            self.reset()

        title = "Manager Override"
        if message:
            title = f"{title} - {message}"

        dlg = ft.AlertDialog(
            modal=True,
            title=ft.Text(title),
            content=ft.Container(
                ft.Column(
                    [
                        ft.Divider(),
                        WuttaLoginForm(
                            self.config,
                            pos=self,
                            perm_required=perm,
                            on_login_success=login_success,
                            on_login_failure=login_failure,
                            on_authz_failure=authz_failure,
                        ),
                    ],
                ),
                height=600,
            ),
            actions=[
                self.make_button("Cancel", on_click=login_cancel, height=80, width=120),
            ],
        )

        # self.page.open(dlg)

        self.page.dialog = dlg
        dlg.open = True
        self.page.update()

    def void_transaction(self, user):
        session = self.app.make_session()
        batch = self.get_current_batch(session)
        user = session.get(user.__class__, user.uuid)
        handler = self.get_batch_handler()

        handler.void_batch(batch, user)

        session.commit()
        session.close()
        self.clear_all()
        self.reset()

    def clear_item_selection(self):
        if self.selected_item:
            self.selected_item.bgcolor = "white"
            self.selected_item.content.refresh()
            self.selected_item = None

    def clear_all(self):
        self.items.controls.clear()

        self.subtotals.spans.clear()
        self.fs_balance.spans.clear()
        self.balances.spans.clear()
        self.totals_row.bgcolor = None

        self.page.session.set("txn_display", None)
        self.page.session.set("cust_uuid", None)
        self.page.session.set("cust_display", None)
        self.informed_refresh()

    def main_submit(self, e=None):
        if self.main_input.value:
            self.attempt_add_product(record_badscan=True)
        self.reset()

    ##############################
    # pos cmd methods
    ##############################

    def cmd(self, cmdname, entry=None, **kwargs):
        """
        Run a POS command.
        """
        meth = getattr(self, f"cmd_{cmdname}", None)
        if meth:
            meth(entry=entry, **kwargs)
        else:
            log.warning(
                "unknown cmd requested: %s, entry=%s, %s", cmdname, repr(entry), kwargs
            )
            self.show_snackbar(f"Unknown command: {cmdname}", bgcolor="yellow")
            self.page.update()

    def cmd_noop(self, entry=None, **kwargs):
        """ """
        self.show_snackbar("Doing nothing", bgcolor="green")
        self.main_input.focus()
        self.page.update()

    def cmd_adjust_price_dwim(self, entry=None, **kwargs):
        enum = self.app.enum

        if not len(self.items.controls):
            self.show_snackbar("There are no line items", bgcolor="yellow")
            self.reset()
            return

        if not self.selected_item:
            self.show_snackbar("Must first select a line item", bgcolor="yellow")
            self.main_input.focus()
            self.page.update()
            return

        row = self.selected_item.data["row"]
        if row.void or row.row_type not in (
            enum.POS_ROW_TYPE_SELL,
            enum.POS_ROW_TYPE_OPEN_RING,
        ):
            self.show_snackbar("This item cannot be adjusted", bgcolor="yellow")
            self.main_input.focus()
            self.page.update()
            return

        self.authorized_action(
            "pos.override_price", self.adjust_price, message="Adjust Price"
        )

    def cmd_context_menu(self, entry=None, **kwargs):
        """
        Swap out which context menu is currently shown.
        """
        spec = self.config.require(f"wuttapos.menus.{entry}.spec")
        factory = self.app.load_object(spec)
        menu = factory(self.config, pos=self)
        self.menu_master.replace_context_menu(menu)

    def cmd_customer_dwim(self, entry=None, **kwargs):

        # prompt user to replace customer if already set
        if self.page.session.get("cust_uuid"):
            self.customer_prompt()

        else:
            value = self.main_input.value
            if value:
                # okay try to set it with given value
                if not self.attempt_set_customer(value):
                    self.customer_lookup(value)

            else:
                # no value provided, so do lookup
                self.customer_lookup()

    def cmd_entry_append(self, entry=None, **kwargs):
        """
        Run a POS command.
        """
        if entry is not None:
            self.main_input.value = f"{self.main_input.value or ''}{entry}"
            self.main_input.focus()
            self.page.update()

    def cmd_entry_submit(self, entry=None, **kwargs):
        self.main_submit()

    def cmd_item_dwim(self, entry=None, **kwargs):

        value = self.main_input.value
        if value:
            if not self.attempt_add_product():
                self.item_lookup(value)

        elif self.selected_item:
            row = self.selected_item.data["row"]
            if row.product_uuid:
                if self.attempt_add_product(uuid=row.product_uuid):
                    self.clear_item_selection()
                    self.page.update()
            else:
                self.item_lookup()

        else:
            self.item_lookup()

    def cmd_item_menu_dept(self, entry=None, **kwargs):
        """
        Show item lookup dialog, restricted to the given department.
        """
        key = entry
        if not key:
            raise ValueError("must specify department key")

        org = self.app.get_org_handler()
        session = self.app.make_session()
        department = org.get_department(session, key)
        session.close()
        if not department:
            raise ValueError(f"department not found: {key}")

        def select(uuid):
            self.attempt_add_product(uuid=uuid)
            dlg.open = False
            self.reset()

        def cancel(e):
            dlg.open = False
            self.reset()

        dlg = ft.AlertDialog(
            title=ft.Text(f"Item from Department: {department.name}"),
            content=WuttaProductLookupByDepartment(
                self.config,
                department,
                page=self.page,
                on_select=select,
                on_cancel=cancel,
            ),
        )

        # self.page.open(dlg)

        self.page.dialog = dlg
        dlg.open = True
        self.page.update()

    def cmd_manager_dwim(self, entry=None, **kwargs):

        def toggle_training(e):
            dlg.open = False
            self.page.update()

            session = self.app.make_session()
            batch = self.get_current_batch(session, create=False)
            session.close()
            if batch:
                self.show_snackbar("TRANSACTION IN PROGRESS")
                self.reset()

            else:
                # nb. do this just in case we must show login dialog
                # cf. https://github.com/flet-dev/flet/issues/1670
                time.sleep(0.1)

                training = self.page.session.get("training")
                toggle = "End" if training else "Start"
                self.authorized_action(
                    "pos.toggle_training",
                    self.toggle_training_mode,
                    message=f"{toggle} Training Mode",
                )

        def cancel(e):
            dlg.open = False
            self.reset()

        font_size = 32
        toggle = "End" if self.page.session.get("training") else "Start"
        dlg = ft.AlertDialog(
            title=ft.Text("Manager Menu"),
            content=ft.Text("What would you like to do?", size=20),
            actions=[
                self.make_button(
                    f"{toggle} Training",
                    font_size=font_size,
                    height=self.default_button_size,
                    width=self.default_button_size * 2.5,
                    bgcolor="yellow",
                    on_click=toggle_training,
                ),
                self.make_button(
                    "Cancel",
                    font_size=font_size,
                    height=self.default_button_size,
                    width=self.default_button_size * 2.5,
                    on_click=cancel,
                ),
            ],
        )

        # self.page.open(dlg)

        self.page.dialog = dlg
        dlg.open = True
        self.page.update()

    def cmd_no_sale_dwim(self, entry=None, **kwargs):

        session = self.app.make_session()
        batch = self.get_current_batch(session, create=False)
        session.close()

        if batch:
            self.show_snackbar("TRANSACTION IN PROGRESS", bgcolor="yellow")
            self.page.update()
            return

        self.kick_drawer()

    def cmd_open_ring_dwim(self, entry=None, **kwargs):

        value = self.main_input.value or None
        if not value:
            self.show_snackbar("Must first enter an amount")
            self.reset()
            return

        amount = self.require_decimal(value)
        if amount is False:
            self.reset()
            return

        def select(uuid):
            session = self.app.make_session()
            user = self.get_current_user(session)
            batch = self.get_current_batch(session, user=user)
            handler = self.get_batch_handler()

            quantity = 1
            if self.set_quantity.data is not None:
                quantity = self.set_quantity.data

            row = handler.add_open_ring(
                batch, uuid, amount, quantity=quantity, user=user
            )
            session.commit()

            session.refresh(row)
            session.expunge(row)
            self.add_row_item(row, scroll=True)
            self.refresh_totals(batch)
            session.close()

            dlg.open = False
            self.reset()

        def cancel(e):
            dlg.open = False
            self.reset(clear_quantity=False)

        dlg = ft.AlertDialog(
            modal=True,
            title=ft.Text(
                f"Department Lookup - for {self.app.render_currency(amount)} OPEN RING"
            ),
            content=WuttaDepartmentLookup(
                self.config, on_select=select, on_cancel=cancel
            ),
        )

        # self.page.open(dlg)

        self.page.dialog = dlg
        dlg.open = True
        self.page.update()

    def cmd_refund_dwim(self, entry=None, **kwargs):
        self.show_snackbar("TODO: handle refund")
        self.page.update()

    def cmd_refresh_txn(self, entry=None, **kwargs):
        session = self.app.make_session()

        if batch := self.get_current_batch(session, create=False):
            self.load_batch(batch)
        else:
            self.page.session.set("txn_display", None)
            self.page.session.set("cust_uuid", None)
            self.page.session.set("cust_display", None)
            self.informed_refresh()

        self.refresh_training()

        # TODO: i think commit() was for when it auto-created the
        # batch, so that can go away now..right?
        # session.commit()
        session.close()
        self.show_snackbar("Transaction refreshed", bgcolor="green")
        self.page.update()

    def cmd_resume_txn(self, entry=None, **kwargs):
        session = self.app.make_session()
        batch = self.get_current_batch(session, create=False)
        session.close()

        # can't resume if txn in progress
        if batch:
            self.show_snackbar("TRANSACTION IN PROGRESS", bgcolor="yellow")
            self.reset()
            return

        def select(uuid):
            model = self.app.model
            session = self.app.make_session()
            user = self.get_current_user(session)
            handler = self.get_batch_handler()

            # TODO: this would need to work differently if suspended
            # txns are kept in a central server DB
            batch = session.get(model.POSBatch, uuid)

            batch = handler.resume_transaction(batch, user)
            session.commit()

            session.refresh(batch)
            self.load_batch(batch)
            session.close()

            dlg.open = False
            self.reset()

        def cancel(e):
            dlg.open = False
            self.reset()

        # prompt to choose txn
        dlg = ft.AlertDialog(
            title=ft.Text("Resume Transaction"),
            content=WuttaTransactionLookup(
                self.config,
                page=self.page,
                mode="resume",
                on_select=select,
                on_cancel=cancel,
            ),
        )

        # self.page.open(dlg)

        self.page.dialog = dlg
        dlg.open = True
        self.page.update()

    def cmd_scroll_down(self, entry=None, **kwargs):

        # select next item, if selection in progress
        if self.selected_item:
            i = self.items.controls.index(self.selected_item)
            if (i + 1) < len(self.items.controls):
                self.items.scroll_to(delta=50, duration=100)
                self.select_txn_item(self.items.controls[i + 1])
                return

        self.items.scroll_to(delta=50, duration=100)
        self.page.update()

    def cmd_scroll_down_page(self, entry=None, **kwargs):
        self.items.scroll_to(delta=500, duration=100)
        self.page.update()

    def cmd_scroll_up(self, entry=None, **kwargs):

        # select previous item, if selection in progress
        if self.selected_item:
            i = self.items.controls.index(self.selected_item)
            if i > 0:
                self.items.scroll_to(delta=-50, duration=100)
                self.select_txn_item(self.items.controls[i - 1])
                return

        self.items.scroll_to(delta=-50, duration=100)
        self.page.update()

    def cmd_scroll_up_page(self, entry=None, **kwargs):
        self.items.scroll_to(delta=-500, duration=100)
        self.page.update()

    def cmd_suspend_txn(self, entry=None, **kwargs):

        session = self.app.make_session()
        batch = self.get_current_batch(session, create=False)
        session.close()

        # nothing to suspend if no txn
        if not batch:
            self.show_snackbar("NO TRANSACTION", bgcolor="yellow")
            self.reset()
            return

        def confirm(e):
            dlg.open = False
            self.page.update()

            # nb. do this just in case we must show login dialog
            # cf. https://github.com/flet-dev/flet/issues/1670
            time.sleep(0.1)

            self.authorized_action(
                "pos.suspend", self.suspend_transaction, message="Suspend Transaction"
            )

        def cancel(e):
            dlg.open = False
            self.reset()

        # prompt to suspend
        dlg = ft.AlertDialog(
            title=ft.Text("Confirm SUSPEND"),
            content=ft.Text("Really SUSPEND transaction?"),
            actions=[
                self.make_button(
                    f"Yes, SUSPEND",
                    font_size=self.default_font_size,
                    height=self.default_button_size,
                    width=self.default_button_size * 3,
                    bgcolor="yellow",
                    on_click=confirm,
                ),
                self.make_button(
                    "Cancel",
                    font_size=self.default_font_size,
                    height=self.default_button_size,
                    width=self.default_button_size * 2.5,
                    on_click=cancel,
                ),
            ],
        )

        # self.page.open(dlg)

        self.page.dialog = dlg
        dlg.open = True
        self.page.update()

    def cmd_tender(self, entry=None, **kwargs):
        # model = self.app.model
        # session = self.app.make_session()
        # handler = self.get_batch_handler()
        # user = self.get_current_user(session)
        # batch = self.get_current_batch(session, user=user, create=False)

        # tender = kwargs.get('tender')
        # if isinstance(tender, model.Tender):
        #     code = tender.code
        # elif tender:
        #     code = tender['code']
        # elif entry:
        #     code = entry
        # if not code:
        #     raise ValueError("must specify tender code")

        self.show_snackbar("TODO: not implemented", bgcolor="yellow")
        self.reset()

        # # nothing to do if no transaction
        # if not batch:
        #     session.close()
        #     self.show_snackbar("NO TRANSACTION", bgcolor='yellow')
        #     self.reset()
        #     return

        # # nothing to do if zero sales
        # if not batch.get_balance():
        #     session.close()
        #     self.show_snackbar("NO SALES", bgcolor='yellow')
        #     self.reset()
        #     return

        # # nothing to do if no amount provided
        # if not self.main_input.value:
        #     session.close()
        #     self.show_snackbar("MUST SPECIFY AMOUNT", bgcolor='yellow')
        #     self.reset()
        #     return

        # # nothing to do if amount not valid
        # amount = self.require_decimal(self.main_input.value)
        # if amount is False:
        #     session.close()
        #     self.reset()
        #     return

        # # do nothing if @ quantity present
        # if self.set_quantity.data:
        #     session.close()
        #     self.show_snackbar(f"QUANTITY NOT ALLOWED FOR TENDER: {self.set_quantity.value}",
        #                        bgcolor='yellow')
        #     self.reset()
        #     return

        # # tender / execute batch
        # try:

        #     # apply tender amount to batch
        #     # nb. this *may* execute the batch!
        #     # nb. we negate the amount supplied by user
        #     rows = handler.apply_tender(batch, user, tender, -amount)

        # except Exception as error:
        #     session.rollback()
        #     log.exception("failed to apply tender '%s' for %s in batch %s",
        #                   code, amount, batch.id_str)
        #     self.show_snackbar(f"ERROR: {error}", bgcolor='red')

        # else:
        #     session.commit()

        #     # update screen to reflect new items/balance
        #     for row in rows:
        #         self.add_row_item(row, scroll=True)
        #     self.refresh_totals(batch)

        #     # executed batch means txn was finalized
        #     if batch.executed:

        #         # look for "change back" row, if found then show alert
        #         last_row = rows[-1]
        #         if last_row.row_type == self.enum.POS_ROW_TYPE_CHANGE_BACK:

        #             def close_bs(e):
        #                 # user dismissed the change back alert; clear screen
        #                 bs.open = False
        #                 bs.update()
        #                 self.clear_all()
        #                 self.reset()

        #             bs = ft.BottomSheet(
        #                 ft.Container(
        #                     ft.Column(
        #                         [
        #                             ft.Text("Change Due", size=24,
        #                                     weight=ft.FontWeight.BOLD),
        #                             ft.Divider(),
        #                             ft.Text("Please give customer their change:",
        #                                     size=20),
        #                             ft.Text(self.app.render_currency(last_row.tender_total),
        #                                     size=32, weight=ft.FontWeight.BOLD),
        #                             ft.Container(
        #                                 content=self.make_button("Dismiss", on_click=close_bs,
        #                                                          height=80, width=120),
        #                                 alignment=ft.alignment.center,
        #                                 expand=1,
        #                             ),
        #                         ],
        #                     ),
        #                     bgcolor='green',
        #                     padding=20,
        #                 ),
        #                 open=True,
        #                 dismissible=False,
        #             )

        #             # show change back alert
        #             # nb. we do *not* clear screen yet
        #             self.page.overlay.append(bs)

        #         else:
        #             # txn finalized but no change back; clear screen
        #             self.clear_all()

        #         # kick drawer if accepting any tender which requires
        #         # that, or if we are giving change back
        #         first_row = rows[0]
        #         if ((first_row.tender and first_row.tender.kick_drawer)
        #             or last_row.row_type == self.enum.POS_ROW_TYPE_CHANGE_BACK):
        #             self.kick_drawer()

        # finally:
        #     session.close()

        # self.reset()

    def cmd_void_dwim(self, entry=None, **kwargs):
        enum = self.app.enum
        session = self.app.make_session()
        batch = self.get_current_batch(session, create=False)
        session.close()

        # nothing to void if no txn
        if not batch:
            self.show_snackbar("NO TRANSACTION", bgcolor="yellow")
            self.reset()
            return

        def confirm(e):
            dlg.open = False
            self.page.update()

            if self.selected_item:

                session = self.app.make_session()
                handler = self.get_batch_handler()
                user = self.get_current_user(session)
                batch = self.get_current_batch(session, user=user)

                # void line
                row = self.selected_item.data["row"]
                if row.void:
                    # cannot void an already void line
                    self.show_snackbar("LINE ALREADY VOID", bgcolor="yellow")

                elif row.row_type not in (
                    enum.POS_ROW_TYPE_SELL,
                    enum.POS_ROW_TYPE_OPEN_RING,
                ):
                    # cannot void line unless of type 'sell'
                    self.show_snackbar("LINE DOES NOT ALLOW VOID", bgcolor="yellow")

                else:
                    # okay, void the line
                    row = session.get(row.__class__, row.uuid)
                    handler.void_row(row, user)
                    session.commit()

                    # refresh display
                    self.selected_item.data["row"] = row
                    self.selected_item.content.row = row
                    self.selected_item.content.refresh()
                    self.clear_item_selection()
                    self.refresh_totals(batch)

                session.close()
                self.reset()

            else:  # void txn

                # nb. do this just in case we must show login dialog
                # cf. https://github.com/flet-dev/flet/issues/1670
                time.sleep(0.1)

                self.authorized_action(
                    "pos.void_txn", self.void_transaction, message="Void Transaction"
                )

        def cancel(e):
            dlg.open = False
            self.reset()

        # prompt to void something
        target = "LINE" if self.selected_item else "TXN"
        dlg = ft.AlertDialog(
            title=ft.Text("Confirm VOID"),
            content=ft.Text(f"Really VOID {target}?"),
            actions=[
                self.make_button(
                    f"VOID {target}",
                    font_size=self.default_font_size,
                    height=self.default_button_size,
                    width=self.default_button_size * 2.5,
                    bgcolor="red",
                    on_click=confirm,
                ),
                self.make_button(
                    "Cancel",
                    font_size=self.default_font_size,
                    height=self.default_button_size,
                    width=self.default_button_size * 2.5,
                    on_click=cancel,
                ),
            ],
        )

        # self.page.open(dlg)

        self.page.dialog = dlg
        dlg.open = True
        self.page.update()



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
See also: :ref:`wuttapos-serve`
"""

import logging

import typer

from wuttapos.cli import wuttapos_typer
from wuttjamaican.util import resource_path


log = logging.getLogger(__name__)


@wuttapos_typer.command()
def serve(ctx: typer.Context):
    """
    Run the WuttaPOS web service
    """
    import flet as ft
    from wuttapos.terminal.app import main

    config = ctx.parent.wutta_config
    kw = {}

    host = config.get("wuttapos.serve.host", default="0.0.0.0")
    kw["host"] = host

    port = config.get_int("wuttapos.serve.port", default=8332)
    kw["port"] = port

    # TODO: we technically "support" this, in that we do pass the
    # value on to Flet, but in practice it does not work right
    path = config.get("wuttapos.serve.path", default="")
    if path:
        path = path.strip("/") + "/"
        kw["name"] = path
        # kw['route_url_strategy'] = 'hash'

    log.info(f"will serve WuttaPOS on http://{host}:{port}/{path}")
    ft.app(
        target=main,
        view=None,
        assets_dir=resource_path("wuttapos.terminal:assets"),
        **kw,
    )



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - keyboard control
"""

import flet as ft


class WuttaKeyboard(ft.Container):

    default_font_size = 20
    default_button_size = 80

    def __init__(self, config, page=None, *args, **kwargs):
        self.on_reset = kwargs.pop("on_reset", None)
        self.on_keypress = kwargs.pop("on_keypress", None)
        self.on_long_backspace = kwargs.pop("on_long_backspace", None)

        super().__init__(*args, **kwargs)

        self.config = config
        self.app = config.get_app()
        self.enum = self.app.enum

        self.caps_lock = False
        self.caps_map = dict([(k, k.upper()) for k in "abcdefghijklmnopqrstuvwxyz"])

        self.shift = False
        self.shift_map = {
            "`": "~",
            "1": "!",
            "2": "@",
            "3": "#",
            "4": "$",
            "5": "%",
            "6": "^",
            "7": "&",
            "8": "*",
            "9": "(",
            "0": ")",
            "-": "_",
            "=": "+",
            "[": "{",
            "]": "}",
            "\\": "|",
            ";": ":",
            "'": '"',
            ",": "<",
            ".": ">",
            "/": "?",
        }

        self.keys = {}

        def make_key(
            key,
            data=None,
            on_click=self.simple_keypress,
            on_long_press=None,
            width=self.default_button_size,
            bgcolor=None,
        ):
            button = ft.Container(
                content=ft.Text(
                    key, size=self.default_font_size, weight=ft.FontWeight.BOLD
                ),
                data=data or key,
                height=self.default_button_size,
                width=width,
                on_click=on_click,
                on_long_press=on_long_press,
                alignment=ft.alignment.center,
                border=ft.border.all(1, "black"),
                border_radius=ft.border_radius.all(5),
                bgcolor=bgcolor,
            )
            self.keys[key] = button
            return button

        def caps_click(e):
            self.update_caps_lock(not self.caps_lock)

        self.caps_key = make_key("CAPS", on_click=caps_click)
        if self.caps_lock:
            self.caps_key.bgcolor = "blue"

        def shift_click(e):
            self.update_shift(not self.shift)

        self.shift_key = make_key("SHIFT", on_click=shift_click)
        if self.shift:
            self.shift_key.bgcolor = "blue"

        rows = [
            [make_key(k) for k in "`1234567890-="]
            + [make_key("⌫", bgcolor="yellow", on_long_press=self.long_backspace)],
            [make_key(k) for k in "qwertyuiop[]\\"],
            [self.caps_key]
            + [make_key(k) for k in "asdfghjkl;'"]
            + [make_key("⏎", bgcolor="blue")],
            [self.shift_key] + [make_key(k) for k in "zxcvbnm,./"],
            [make_key("SPACE", width=self.default_button_size * 5)],
        ]

        rows = [
            ft.Row(controls, alignment=ft.MainAxisAlignment.CENTER) for controls in rows
        ]

        self.content = ft.Column(rows)

    def informed_refresh(self, **kwargs):
        pass

    def reset(self, e=None):
        if self.on_reset:
            self.on_reset(e=e)

    def update_caps_lock(self, caps_lock):
        self.caps_lock = caps_lock

        if self.caps_lock:
            self.caps_key.bgcolor = "blue"
        else:
            self.caps_key.bgcolor = None

        for key, button in self.keys.items():
            if key in self.caps_map:
                if self.caps_lock or self.shift:
                    button.content.value = self.caps_map[key]
                else:
                    button.content.value = key

        self.update()

    def update_shift(self, shift):
        self.shift = shift

        if self.shift:
            self.shift_key.bgcolor = "blue"
        else:
            self.shift_key.bgcolor = None

        for key, button in self.keys.items():
            if key in self.caps_map:
                if self.shift or self.caps_lock:
                    button.content.value = self.caps_map[key]
                else:
                    button.content.value = key
            elif key in self.shift_map:
                if self.shift:
                    button.content.value = self.shift_map[key]
                else:
                    button.content.value = key

        self.update()

    def simple_keypress(self, e):

        # maybe inform parent
        if self.on_keypress:
            key = e.control.content.value

            # avoid callback for certain keys
            if key not in ("CAPS", "SHIFT"):

                # translate certain keys
                if key == "SPACE":
                    key = " "

                # let 'em know
                self.on_keypress(key)

        # turn off shift key if set
        if self.shift:
            self.update_shift(False)

    def long_backspace(self, e):
        if self.on_long_backspace:
            self.on_long_backspace()



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - flet views (base class)
"""

import os

from wuttjamaican.util import resource_path

import flet as ft

from wuttapos.terminal.controls.header import WuttaHeader
from wuttapos.terminal.controls.buttons import make_button
from wuttapos.terminal.util import show_snackbar, get_pos_batch_handler


class WuttaView(ft.View):
    """
    Base class for all Flet views used in WuttaPOS
    """

    def __init__(self, config, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config = config
        self.app = self.config.get_app()

        controls = self.build_controls()
        self.controls = [
            WuttaViewContainer(
                self.config, content=ft.Column(controls=controls), expand=True
            ),
        ]

    def build_controls(self):
        return [self.build_header()]

    def build_header(self):
        handler = self.get_batch_handler()
        self.header = WuttaHeader(
            self.config, on_reset=self.reset, terminal_id=handler.get_terminal_id()
        )
        return self.header

    def get_batch_handler(self):
        return get_pos_batch_handler(self.config)

    def make_button(self, *args, **kwargs):
        return make_button(*args, **kwargs)

    def reset(self, *args, **kwargs):
        pass

    def make_logo_image(self, **kwargs):

        # we have a default header logo, but prefer custom if present
        custom = resource_path("wuttapos.terminal:assets/custom_header_logo.png")
        if os.path.exists(custom):
            logo = "/custom_header_logo.png"
        else:
            logo = "/header_logo.png"

        # but config can override in any case
        logo = self.config.get("wuttapos.header.logo", default=logo)

        kwargs.setdefault("height", 100)
        return ft.Image(src=logo, **kwargs)

    def show_snackbar(self, text, bgcolor="yellow"):
        show_snackbar(self.page, text, bgcolor=bgcolor)


class WuttaViewContainer(ft.Container):
    """
    Main container class to wrap all controls for a view.  Used for
    displaying background image etc.
    """

    def __init__(self, config, *args, **kwargs):
        self.config = config

        # # add testing watermark when not in production
        # if "image_src" not in kwargs and not self.config.production():
        #     kwargs["image_src"] = "/testing.png"
        #     kwargs.setdefault("image_repeat", ft.ImageRepeat.REPEAT)

        super().__init__(*args, **kwargs)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - timestamp control
"""

import asyncio
import datetime
import threading
import time

import flet as ft


class WuttaTimestamp(ft.Text):

    def __init__(self, config, page=None, *args, **kwargs):
        self.on_reset = kwargs.pop("on_reset", None)

        super().__init__(*args, **kwargs)

        self.config = config
        self.app = self.config.get_app()

        self.value = self.render_time(datetime.datetime.now())

    def did_mount(self):
        self.running = True
        if hasattr(self.page, "run_task"):
            self.page.run_task(self.update_display)
        else:
            # nb. daemonized thread should be stopped when app exits
            # cf. https://docs.python.org/3/library/threading.html#thread-objects
            thread = threading.Thread(target=self.update_display_blocking, daemon=True)
            thread.start()

    def will_unmount(self):
        self.running = False

    def render_time(self, value):
        return value.strftime("%a %d %b %Y - %I:%M:%S %p")

    async def update_display(self):
        while self.running:
            self.value = self.render_time(datetime.datetime.now())
            self.update()
            await asyncio.sleep(0.5)

    def update_display_blocking(self):
        while self.running:
            # self.value = self.render_time(self.app.localtime())
            self.value = self.render_time(datetime.datetime.now())
            self.update()
            time.sleep(0.5)



"""add Products

Revision ID: 6f02663c2220
Revises: 148d701e4ac7
Create Date: 2026-01-01 18:18:22.958598

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "6f02663c2220"
down_revision: Union[str, None] = "148d701e4ac7"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # product
    op.create_table(
        "product",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("product_id", sa.String(length=20), nullable=False),
        sa.Column("brand_name", sa.String(length=100), nullable=True),
        sa.Column("description", sa.String(length=255), nullable=False),
        sa.Column("size", sa.String(length=30), nullable=True),
        sa.Column("sold_by_weight", sa.Boolean(), nullable=False),
        sa.Column("department_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("special_order", sa.Boolean(), nullable=True),
        sa.Column("case_size", sa.Numeric(precision=9, scale=4), nullable=True),
        sa.Column("unit_cost", sa.Numeric(precision=9, scale=5), nullable=True),
        sa.Column("unit_price_reg", sa.Numeric(precision=8, scale=3), nullable=True),
        sa.Column("notes", sa.Text(), nullable=True),
        sa.ForeignKeyConstraint(
            ["department_uuid"],
            ["department.uuid"],
            name=op.f("fk_product_department_uuid_department"),
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_product")),
    )
    op.create_table(
        "product_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "product_id", sa.String(length=20), autoincrement=False, nullable=True
        ),
        sa.Column(
            "brand_name", sa.String(length=100), autoincrement=False, nullable=True
        ),
        sa.Column(
            "description", sa.String(length=255), autoincrement=False, nullable=True
        ),
        sa.Column("size", sa.String(length=30), autoincrement=False, nullable=True),
        sa.Column("sold_by_weight", sa.Boolean(), autoincrement=False, nullable=True),
        sa.Column(
            "department_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("special_order", sa.Boolean(), autoincrement=False, nullable=True),
        sa.Column(
            "case_size",
            sa.Numeric(precision=9, scale=4),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "unit_cost",
            sa.Numeric(precision=9, scale=5),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "unit_price_reg",
            sa.Numeric(precision=8, scale=3),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("notes", sa.Text(), autoincrement=False, nullable=True),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_product_version")
        ),
    )
    op.create_index(
        op.f("ix_product_version_end_transaction_id"),
        "product_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_product_version_operation_type"),
        "product_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_product_version_transaction_id"),
        "product_version",
        ["transaction_id"],
        unique=False,
    )

    # product_inventory
    op.create_table(
        "product_inventory",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("product_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("on_hand", sa.Numeric(precision=9, scale=4), nullable=True),
        sa.Column("on_order", sa.Numeric(precision=9, scale=4), nullable=True),
        sa.ForeignKeyConstraint(
            ["product_uuid"],
            ["product.uuid"],
            name=op.f("fk_product_inventory_product_uuid_product"),
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_product_inventory")),
    )

    # inventory_adjustment_type
    op.create_table(
        "inventory_adjustment_type",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("type_code", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=100), nullable=False),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_inventory_adjustment_type")),
        sa.UniqueConstraint(
            "type_code", name=op.f("uq_inventory_adjustment_type_type_code")
        ),
    )
    op.create_table(
        "inventory_adjustment_type_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column("type_code", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column("name", sa.String(length=100), autoincrement=False, nullable=True),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_inventory_adjustment_type_version")
        ),
    )
    op.create_index(
        op.f("ix_inventory_adjustment_type_version_end_transaction_id"),
        "inventory_adjustment_type_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_inventory_adjustment_type_version_operation_type"),
        "inventory_adjustment_type_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_inventory_adjustment_type_version_transaction_id"),
        "inventory_adjustment_type_version",
        ["transaction_id"],
        unique=False,
    )

    # inventory_adjustment
    op.create_table(
        "inventory_adjustment",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("inventory_uuid", wuttjamaican.db.util.UUID(), nullable=True),
        sa.Column("adjusted", sa.DateTime(), nullable=False),
        sa.Column("effective_date", sa.Date(), nullable=False),
        sa.Column("adjustment_type_uuid", wuttjamaican.db.util.UUID(), nullable=True),
        sa.Column("amount", sa.Numeric(precision=9, scale=4), nullable=False),
        sa.Column("source", sa.String(length=100), nullable=True),
        sa.ForeignKeyConstraint(
            ["adjustment_type_uuid"],
            ["inventory_adjustment_type.uuid"],
            name=op.f(
                "fk_inventory_adjustment_adjustment_type_uuid_inventory_adjustment_type"
            ),
        ),
        sa.ForeignKeyConstraint(
            ["inventory_uuid"],
            ["product_inventory.uuid"],
            name=op.f("fk_inventory_adjustment_inventory_uuid_product_inventory"),
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_inventory_adjustment")),
    )


def downgrade() -> None:

    # inventory_adjustment
    op.drop_table("inventory_adjustment")

    # inventory_adjustment_type
    op.drop_index(
        op.f("ix_inventory_adjustment_type_version_transaction_id"),
        table_name="inventory_adjustment_type_version",
    )
    op.drop_index(
        op.f("ix_inventory_adjustment_type_version_operation_type"),
        table_name="inventory_adjustment_type_version",
    )
    op.drop_index(
        op.f("ix_inventory_adjustment_type_version_end_transaction_id"),
        table_name="inventory_adjustment_type_version",
    )
    op.drop_table("inventory_adjustment_type_version")
    op.drop_table("inventory_adjustment_type")

    # product_inventory
    op.drop_table("product_inventory")

    # product
    op.drop_index(
        op.f("ix_product_version_transaction_id"), table_name="product_version"
    )
    op.drop_index(
        op.f("ix_product_version_operation_type"), table_name="product_version"
    )
    op.drop_index(
        op.f("ix_product_version_end_transaction_id"), table_name="product_version"
    )
    op.drop_table("product_version")
    op.drop_table("product")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - menus master control
"""

import flet as ft

from wuttapos.terminal.controls.menus.tenkey import WuttaTenkeyMenu
from wuttapos.terminal.controls.menus.meta import WuttaMetaMenu
from wuttapos.terminal.controls.menus.suspend import WuttaSuspendMenu


class WuttaMenuMaster(ft.Column):
    """
    Base class and default implementation for "buttons master"
    control.  This represents the overall button area in POS view.
    """

    def __init__(self, config, pos=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config = config
        self.app = self.config.get_app()
        self.pos = pos
        self.controls = self.build_controls()

    def build_controls(self):
        self.tenkey_menu = self.build_tenkey_menu()
        self.meta_menu = self.build_meta_menu()
        self.context_menu = self.build_context_menu()
        self.suspend_menu = self.build_suspend_menu()

        return [
            ft.Row(
                [
                    self.tenkey_menu,
                    self.meta_menu,
                ],
            ),
            ft.Row(
                [
                    self.context_menu,
                    self.suspend_menu,
                ],
                vertical_alignment=ft.CrossAxisAlignment.START,
            ),
        ]

    ##############################
    # tenkey
    ##############################

    def build_tenkey_menu(self):
        return WuttaTenkeyMenu(
            self.config,
            pos=self.pos,
            on_char=self.tenkey_char,
            on_enter=self.tenkey_enter,
            on_up_click=self.tenkey_up_click,
            on_up_longpress=self.tenkey_up_longpress,
            on_down_click=self.tenkey_down_click,
            on_down_longpress=self.tenkey_down_longpress,
        )

    def tenkey_char(self, key):
        self.pos.cmd("entry_append", key)

    def tenkey_enter(self, e):
        self.pos.cmd("entry_submit")

    def tenkey_up_click(self, e):
        self.pos.cmd("scroll_up")

    def tenkey_up_longpress(self, e):
        self.pos.cmd("scroll_up_page")

    def tenkey_down_click(self, e):
        self.pos.cmd("scroll_down")

    def tenkey_down_longpress(self, e):
        self.pos.cmd("scroll_down_page")

    ##############################
    # meta
    ##############################

    def build_meta_menu(self):
        return WuttaMetaMenu(self.config, pos=self.pos)

    ##############################
    # context
    ##############################

    def build_context_menu(self):
        spec = self.config.get(
            "wuttapos.menus.context.spec",
            default="wuttapos.terminal.controls.menus.context:WuttaContextMenu",
        )
        factory = self.app.load_object(spec)
        return factory(self.config, pos=self.pos)

    def replace_context_menu(self, menu):
        controls = menu.build_controls()
        self.context_menu.content.controls = controls
        self.update()

    ##############################
    # suspend
    ##############################

    def build_suspend_menu(self):
        return WuttaSuspendMenu(self.config, pos=self.pos)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - button menus
"""

import flet as ft

from wuttapos.terminal.controls.buttons import make_button, WuttaButtonRow


class WuttaMenu(ft.Container):
    """
    Base class for button menu controls.
    """

    # TODO: should be configurable somehow
    default_button_size = 100
    default_font_size = 40

    def __init__(self, config, pos=None, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self.config = config
        self.app = self.config.get_app()
        self.pos = pos

        self.content = ft.Column(controls=self.build_controls(), spacing=0)

    def make_button(self, *args, **kwargs):
        kwargs.setdefault("font_size", self.default_font_size)
        kwargs.setdefault("height", self.default_button_size)
        kwargs.setdefault("width", self.default_button_size * 2)
        kwargs.setdefault("pos", self.pos)
        return make_button(*args, **kwargs)

    def make_button_row(self, *args, **kwargs):
        return WuttaButtonRow(*args, **kwargs)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Form schema types
"""

from wuttaweb.forms.schema import ObjectRef


class StoreRef(ObjectRef):
    """
    Schema type for a
    :class:`~wuttapos.db.model.stores.Store` reference field.

    This is a subclass of
    :class:`~wuttaweb:wuttaweb.forms.schema.ObjectRef`.
    """

    @property
    def model_class(self):
        model = self.app.model
        return model.Store

    def sort_query(self, query):
        return query.order_by(self.model_class.name)

    def get_object_url(self, obj):
        store = obj
        return self.request.route_url("stores.view", uuid=store.uuid)


class TerminalRef(ObjectRef):
    """
    Schema type for a
    :class:`~wuttapos.db.model.terminals.Terminal` reference field.

    This is a subclass of
    :class:`~wuttaweb:wuttaweb.forms.schema.ObjectRef`.
    """

    @property
    def model_class(self):
        model = self.app.model
        return model.Terminal

    def sort_query(self, query):
        return query.order_by(self.model_class.name)

    def get_object_url(self, obj):
        terminal = obj
        return self.request.route_url("terminals.view", uuid=terminal.uuid)


class DepartmentRef(ObjectRef):
    """
    Schema type for a
    :class:`~wuttapos.db.model.departments.Department` reference field.

    This is a subclass of
    :class:`~wuttaweb:wuttaweb.forms.schema.ObjectRef`.
    """

    @property
    def model_class(self):
        model = self.app.model
        return model.Department

    def sort_query(self, query):
        return query.order_by(self.model_class.name)

    def get_object_url(self, obj):
        department = obj
        return self.request.route_url("departments.view", uuid=department.uuid)


class EmployeeRef(ObjectRef):
    """
    Schema type for a
    :class:`~wuttapos.db.model.employees.Employee` reference field.

    This is a subclass of
    :class:`~wuttaweb:wuttaweb.forms.schema.ObjectRef`.
    """

    @property
    def model_class(self):
        model = self.app.model
        return model.Employee

    def sort_query(self, query):
        return query.order_by(self.model_class.name)

    def get_object_url(self, obj):
        employee = obj
        return self.request.route_url("employees.view", uuid=employee.uuid)


class CustomerRef(ObjectRef):
    """
    Schema type for a
    :class:`~wuttapos.db.model.customers.Customer` reference field.

    This is a subclass of
    :class:`~wuttaweb:wuttaweb.forms.schema.ObjectRef`.
    """

    @property
    def model_class(self):
        model = self.app.model
        return model.Customer

    def sort_query(self, query):
        return query.order_by(self.model_class.name)

    def get_object_url(self, obj):
        customer = obj
        return self.request.route_url("customers.view", uuid=customer.uuid)


class ProductRef(ObjectRef):
    """
    Schema type for a
    :class:`~wuttapos.db.model.products.Product` reference field.

    This is a subclass of
    :class:`~wuttaweb:wuttaweb.forms.schema.ObjectRef`.
    """

    @property
    def model_class(self):
        model = self.app.model
        return model.Product

    def sort_query(self, query):
        return query.order_by(self.model_class.description)

    def get_object_url(self, obj):
        product = obj
        return self.request.route_url("products.view", uuid=product.uuid)


class InventoryAdjustmentTypeRef(ObjectRef):
    """
    Schema type for a
    :class:`~wuttapos.db.model.products.InventoryAdjustmentType`
    reference field.

    This is a subclass of
    :class:`~wuttaweb:wuttaweb.forms.schema.ObjectRef`.
    """

    @property
    def model_class(self):
        model = self.app.model
        return model.InventoryAdjustmentType

    def sort_query(self, query):
        return query.order_by(self.model_class.name)

    def get_object_url(self, obj):
        adjustment_type = obj
        return self.request.route_url(
            "inventory_adjustment_types.view", uuid=adjustment_type.uuid
        )



# -*- coding: utf-8; -*-
"""
Model definition for Taxes
"""

import sqlalchemy as sa
from sqlalchemy import orm

from wuttjamaican.db import model


class Tax(model.Base):
    """
    Represents a type/rate of sales tax to track.
    """

    __tablename__ = "tax"
    __versioned__ = {}
    __wutta_hint__ = {
        "model_title": "Tax",
        "model_title_plural": "Taxes",
    }

    uuid = model.uuid_column()

    tax_id = sa.Column(
        sa.String(length=20),
        nullable=False,
        doc="""
        Unique identifier for this tax rate.
        """,
    )

    name = sa.Column(
        sa.String(length=50),
        nullable=False,
        doc="""
        Name for the tax rate.
        """,
    )

    rate = sa.Column(
        sa.Numeric(precision=7, scale=5),
        nullable=False,
        doc="""
        Percentage rate for the tax, e.g. 8.25.
        """,
    )

    def __str__(self):
        return self.name or ""



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Install handler for WuttaPOS
"""

import subprocess
import sys

from wuttjamaican import install as base


class InstallHandler(base.InstallHandler):
    """
    Custom install handler for WuttaPOS
    """

    template_paths = ["wuttapos:installer-templates"]

    def do_install_steps(self):

        # prompt for install type first
        self.get_install_type()

        # then everything else
        super().do_install_steps()

    def get_install_type(self):

        # prompt user
        install_type = None
        while install_type not in ("server", "terminal"):
            install_type = self.prompt_generic("install type (server/terminal)")

        # remember the answer
        self.install_type = install_type

        if self.install_type != "server":
            self.rprint(
                "[bold red]sorry, terminal install is not yet implemented[/bold red]\n"
            )
            sys.exit(1)

        # install dependencies
        if self.install_type == "server":
            self.install_server_deps()

    def install_server_deps(self):
        subprocess.check_call(
            [sys.executable, "-m", "pip", "install", "WuttaPOS[server]"]
        )

    def make_template_context(self, dbinfo, **kwargs):
        context = super().make_template_context(dbinfo, **kwargs)
        context["install_type"] = self.install_type
        return context



# -*- coding: utf-8; -*-
"""
Model definition for Stores
"""

import sqlalchemy as sa
from sqlalchemy import orm

from wuttjamaican.db import model


class Store(model.Base):
    """
    Represents a single location, physical or virtual, where sales happen.
    """

    __tablename__ = "store"
    __versioned__ = {}
    __wutta_hint__ = {
        "model_title": "Store",
        "model_title_plural": "Stores",
    }

    uuid = model.uuid_column()

    store_id = sa.Column(
        sa.String(length=20),
        nullable=False,
        doc="""
        Unique identifier for the store.
        """,
    )

    name = sa.Column(
        sa.String(length=100),
        nullable=False,
        doc="""
        Name for the store.
        """,
    )

    active = sa.Column(
        sa.Boolean(),
        nullable=False,
        default=True,
        doc="""
        Indicates the store is currently active.
        """,
    )

    def __str__(self):
        return self.name or ""



"""add Terminals

Revision ID: 653b7d27c709
Revises: 3f548013be91
Create Date: 2026-01-02 11:08:43.118117

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "653b7d27c709"
down_revision: Union[str, None] = "3f548013be91"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # terminal
    op.create_table(
        "terminal",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("terminal_id", sa.String(length=20), nullable=False),
        sa.Column("name", sa.String(length=50), nullable=False),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_terminal")),
    )
    op.create_table(
        "terminal_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "terminal_id", sa.String(length=20), autoincrement=False, nullable=True
        ),
        sa.Column("name", sa.String(length=50), autoincrement=False, nullable=True),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_terminal_version")
        ),
    )
    op.create_index(
        op.f("ix_terminal_version_end_transaction_id"),
        "terminal_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_terminal_version_operation_type"),
        "terminal_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_terminal_version_transaction_id"),
        "terminal_version",
        ["transaction_id"],
        unique=False,
    )


def downgrade() -> None:

    # terminal
    op.drop_index(
        op.f("ix_terminal_version_transaction_id"), table_name="terminal_version"
    )
    op.drop_index(
        op.f("ix_terminal_version_operation_type"), table_name="terminal_version"
    )
    op.drop_index(
        op.f("ix_terminal_version_end_transaction_id"), table_name="terminal_version"
    )
    op.drop_table("terminal_version")
    op.drop_table("terminal")



# -*- coding: utf-8; -*-
"""
Master view for Employees
"""

from wuttaweb.views import MasterView
from wuttaweb.forms.schema import PersonRef

from wuttapos.db.model import Employee


class EmployeeView(MasterView):
    """
    Master view for Employees
    """

    model_class = Employee
    model_title = "Employee"
    model_title_plural = "Employees"

    route_prefix = "employees"
    url_prefix = "/employees"

    creatable = True
    editable = True
    deletable = True

    grid_columns = [
        "person",
        "name",
        "public_name",
        "active",
    ]

    form_fields = [
        "person",
        "name",
        "public_name",
        "active",
    ]

    def configure_form(self, form):
        f = form
        super().configure_form(f)

        # person
        f.set_node("person", PersonRef(self.request))


def defaults(config, **kwargs):
    base = globals()

    EmployeeView = kwargs.get("EmployeeView", base["EmployeeView"])
    EmployeeView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - header control
"""

import datetime

import flet as ft

import wuttapos
from .timestamp import WuttaTimestamp
from .feedback import WuttaFeedback
from wuttapos.terminal.controls.buttons import make_button


class WuttaHeader(ft.Stack):

    def __init__(self, config, page=None, *args, **kwargs):
        self.terminal_id = kwargs.pop("terminal_id", None)
        self.on_reset = kwargs.pop("on_reset", None)

        super().__init__(*args, **kwargs)

        self.config = config
        self.app = config.get_app()
        self.enum = self.app.enum

        self.txn_display = ft.Text("Txn: N", weight=ft.FontWeight.BOLD, size=20)
        self.cust_display = ft.Text("Cust: N", weight=ft.FontWeight.BOLD, size=20)

        self.training_mode = ft.Text(size=40, weight=ft.FontWeight.BOLD)

        self.user_display = ft.Text("User: N", weight=ft.FontWeight.BOLD, size=20)
        self.logout_button = ft.OutlinedButton(
            "Logout", on_click=self.logout_click, visible=False
        )
        self.logout_divider = ft.VerticalDivider(visible=False)
        self.title_button = ft.FilledButton(
            self.app.get_title(), on_click=self.title_click
        )

        terminal_style = ft.TextStyle(size=20, weight=ft.FontWeight.BOLD)
        if not self.terminal_id:
            terminal_style.bgcolor = "red"
            terminal_style.color = "white"

        self.controls = [
            ft.Container(
                content=ft.Row(
                    [
                        ft.Container(
                            content=self.training_mode,
                            bgcolor="yellow",
                        ),
                    ],
                    alignment=ft.MainAxisAlignment.CENTER,
                ),
            ),
            ft.Row(
                [
                    ft.Row(
                        [
                            self.txn_display,
                            ft.VerticalDivider(),
                            self.cust_display,
                            ft.VerticalDivider(),
                            WuttaTimestamp(
                                self.config, weight=ft.FontWeight.BOLD, size=20
                            ),
                        ],
                    ),
                    ft.Row(
                        [
                            self.user_display,
                            ft.VerticalDivider(),
                            self.logout_button,
                            self.logout_divider,
                            ft.Text(
                                spans=[
                                    ft.TextSpan(
                                        style=terminal_style,
                                        text=f"Term: {self.terminal_id or '??'}",
                                    ),
                                ],
                            ),
                            ft.VerticalDivider(),
                            self.title_button,
                        ],
                    ),
                ],
                alignment=ft.MainAxisAlignment.SPACE_BETWEEN,
            ),
        ]

    def reset(self, e=None):
        if self.on_reset:
            self.on_reset(e=e)

    def did_mount(self):
        self.informed_refresh()

    def informed_refresh(self):
        self.update_txn_display()
        self.update_cust_display()
        self.update_user_display()
        self.update_training_display()
        self.update()

    def update_txn_display(self):
        txn_display = None
        if self.page:
            txn_display = self.page.session.get("txn_display")
        self.txn_display.value = f"Txn: {txn_display or 'N'}"

    def update_cust_display(self):
        cust_display = None
        if self.page:
            cust_display = self.page.session.get("cust_display")
        self.cust_display.value = f"Cust: {cust_display or 'N'}"

    def update_training_display(self):
        if self.page.session.get("training"):
            self.training_mode.value = "  TRAINING MODE  "
        else:
            self.training_mode.value = ""

    def update_user_display(self):
        user_display = None
        if self.page:
            user_display = self.page.session.get("user_display")
        self.user_display.value = f"User: {user_display or 'N'}"

        if self.page and self.page.session.get("user_uuid"):
            self.logout_button.visible = True
            self.logout_divider.visible = True

    def logout_click(self, e):

        # TODO: hacky but works for now
        if not self.config.production():
            self.page.client_storage.set("user_uuid", "")

        self.page.session.clear()
        self.page.go("/login")

    def title_click(self, e):
        title = self.app.get_title()

        year = self.app.localtime().year
        if year > 2026:
            year_range = f"2026 - {year}"
        else:
            year_range = year

        license = f"""\
WuttaPOS -- Point of Sale system based on Wutta Framework
Copyright © {year_range} Lance Edgar

WuttaPOS is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
"""

        buttons = []

        user_uuid = self.page.session.get("user_uuid")
        if user_uuid:

            session = self.app.make_session()
            model = self.app.model
            user = session.get(model.User, user_uuid)
            auth = self.app.get_auth_handler()
            has_perm = auth.has_permission(session, user, "pos.test_error")
            session.close()
            if has_perm:
                test_error = make_button(
                    "TEST ERROR",
                    font_size=24,
                    height=60,
                    width=60 * 3,
                    bgcolor="red",
                    on_click=self.test_error_click,
                )
                buttons.append(test_error)

            feedback = WuttaFeedback(
                self.config, page=self.page, on_send=self.reset, on_cancel=self.reset
            )
            buttons.append(feedback)

        self.dlg = ft.AlertDialog(
            title=ft.Text(title),
            content=ft.Container(
                content=ft.Column(
                    [
                        ft.Divider(),
                        ft.Text(f"{title} v{wuttapos.__version__}"),
                        ft.Divider(),
                        ft.Text(license),
                        ft.Container(
                            content=ft.Row(
                                controls=buttons,
                                alignment=ft.MainAxisAlignment.CENTER,
                            ),
                            alignment=ft.alignment.center,
                            expand=True,
                        ),
                    ],
                    expand=True,
                ),
                height=600,
            ),
            actions=[
                ft.Row(
                    [
                        ft.Container(
                            content=ft.Text(
                                "Close", size=20, weight=ft.FontWeight.BOLD
                            ),
                            height=60,
                            width=60 * 2.5,
                            alignment=ft.alignment.center,
                            border=ft.border.all(1, "black"),
                            border_radius=ft.border_radius.all(5),
                            on_click=self.close_dlg,
                        ),
                    ],
                    alignment=ft.MainAxisAlignment.CENTER,
                ),
            ],
        )

        # self.page.open(self.dlg)

        self.page.dialog = self.dlg
        self.dlg.open = True
        self.page.update()

    def test_error_click(self, e):

        # first get the dialog out of the way
        self.dlg.open = False
        self.reset()
        self.page.update()

        raise RuntimeError("FAKE ERROR - to test error handling")

    def close_dlg(self, e):
        self.dlg.open = False
        self.reset()
        self.page.update()



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - buttons
"""

import flet as ft


def make_button(text, font_size=24, font_bold=True, font_weight=None, **kwargs):
    """
    Generic function for making a button.
    """
    if "content" not in kwargs:
        if not font_weight and font_bold:
            font_weight = ft.FontWeight.BOLD
        text = ft.Text(
            text, size=font_size, weight=font_weight, text_align=ft.TextAlign.CENTER
        )
        kwargs["content"] = text

    return WuttaButton(**kwargs)


class WuttaButton(ft.Container):
    """
    Base class for buttons to be shown in the POS menu etc.
    """

    def __init__(
        self,
        pos=None,
        pos_cmd=None,
        pos_cmd_entry=None,
        pos_cmd_kwargs={},
        *args,
        **kwargs
    ):
        kwargs.setdefault("alignment", ft.alignment.center)
        kwargs.setdefault("border", ft.border.all(1, "black"))
        kwargs.setdefault("border_radius", ft.border_radius.all(5))
        super().__init__(*args, **kwargs)

        self.pos = pos
        self.pos_cmd = pos_cmd
        self.pos_cmd_entry = pos_cmd_entry
        self.pos_cmd_kwargs = pos_cmd_kwargs

        if not kwargs.get("on_click") and self.pos and self.pos_cmd:
            self.on_click = self.handle_click

    def handle_click(self, e):
        self.pos.cmd(self.pos_cmd, entry=self.pos_cmd_entry, **self.pos_cmd_kwargs)


class WuttaButtonRow(ft.Row):
    """
    Base class for a row of buttons
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("spacing", 0)
        super().__init__(*args, **kwargs)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - login form control
"""

import logging

import flet as ft

from wuttapos.terminal.controls.buttons import make_button
from wuttapos.terminal.controls.keyboard import WuttaKeyboard
from wuttapos.terminal.controls.menus.tenkey import WuttaTenkeyMenu


log = logging.getLogger(__name__)


class WuttaLoginForm(ft.Column):

    def __init__(self, config, page=None, pos=None, *args, **kwargs):
        self.on_reset = kwargs.pop("on_reset", None)

        # permission to be checked for login to succeed
        self.perm_required = kwargs.pop("perm_required", "pos.ring_sales")

        # may or may not show the username field
        # nb. must set this before normal __init__
        if "show_username" in kwargs:
            self.show_username = kwargs.pop("show_username")
        else:
            self.show_username = config.get_bool(
                "wuttapos.login.show_username", default=True
            )

        # may or may not show 10-key menu instead of full keyboard
        if "use_tenkey" in kwargs:
            self.use_tenkey = kwargs.pop("use_tenkey")
        else:
            self.use_tenkey = config.get_bool(
                "wuttapos.login.use_tenkey", default=False
            )

        self.on_login_failure = kwargs.pop("on_login_failure", None)
        self.on_authz_failure = kwargs.pop("on_authz_failure", None)
        self.on_login_success = kwargs.pop("on_login_success", None)

        super().__init__(*args, **kwargs)

        self.config = config
        self.app = config.get_app()
        self.enum = self.app.enum
        self.pos = pos

        # track which login input has focus
        self.focused = None

        login_form = self.build_login_form()
        self.expand = True
        self.alignment = ft.MainAxisAlignment.CENTER

        if self.use_tenkey:
            self.controls = [
                ft.Row(
                    [
                        login_form,
                        ft.VerticalDivider(),
                        WuttaTenkeyMenu(
                            self.config,
                            pos=self.pos,
                            simple=True,
                            on_char=self.tenkey_char,
                            on_enter=self.tenkey_enter,
                        ),
                    ],
                    alignment=ft.MainAxisAlignment.CENTER,
                ),
            ]

        else:  # full keyboard
            self.controls = [
                login_form,
                ft.Row(),
                ft.Row(),
                ft.Row(),
                WuttaKeyboard(
                    self.config,
                    on_keypress=self.keyboard_keypress,
                    on_long_backspace=self.keyboard_long_backspace,
                ),
            ]

    def informed_refresh(self, **kwargs):
        pass

    def reset(self, e=None):
        if self.on_reset:
            self.on_reset(e=e)

    def build_login_form(self):
        form_fields = []

        self.password = ft.TextField(
            label="Password",
            width=200,
            password=True,
            on_submit=self.password_submit,
            on_focus=self.password_focus,
            autofocus=not self.show_username,
        )
        self.focused = self.password

        if self.show_username:
            self.username = ft.TextField(
                label="Login",
                width=200,
                on_submit=self.username_submit,
                on_focus=self.username_focus,
                autofocus=True,
            )
            form_fields.append(self.username)
            self.focused = self.username

        form_fields.append(self.password)

        login_button = make_button(
            "Login",
            height=60,
            width=60 * 2.5,
            bgcolor="blue",
            on_click=self.attempt_login,
        )

        reset_button = make_button(
            "Clear", height=60, width=60 * 2.5, on_click=self.clear_login
        )

        if self.use_tenkey:
            form_fields.extend(
                [
                    ft.Row(),
                    ft.Row(),
                    ft.Row(
                        [
                            reset_button,
                            login_button,
                        ],
                    ),
                ]
            )
            return ft.Column(
                controls=form_fields,
                horizontal_alignment=ft.CrossAxisAlignment.CENTER,
            )

        else:  # full keyboard
            form_fields.extend(
                [
                    login_button,
                    reset_button,
                ]
            )
            return ft.Row(
                controls=form_fields,
                alignment=ft.MainAxisAlignment.CENTER,
            )

    def keyboard_keypress(self, key):
        assert self.focused

        if key == "⏎":  # ENTER

            # attempt to submit the login form..
            if self.show_username and self.focused is self.username:
                self.username_submit()
            else:
                if self.password_submit():

                    # here the login has totally worked, which means
                    # this form has fulfilled its purpose.  hence must
                    # exit early to avoid update() in case we are to
                    # be redirected etc.  otherwise may get errors
                    # trying to update controls which have already
                    # been dropped from the page..
                    return

        elif key == "⌫":
            self.focused.value = self.focused.value[:-1]

        else:
            self.focused.value += key

        self.focused.focus()
        self.update()

    def keyboard_long_backspace(self):
        assert self.focused
        self.focused.value = ""
        self.focused.focus()
        self.update()

    def tenkey_char(self, key):
        if key == "@":
            return

        self.focused.value = f"{self.focused.value or ''}{key}"
        self.update()

    def tenkey_enter(self, e):
        if self.show_username and self.focused is self.username:
            self.username_submit(e)
            self.update()
        else:
            if not self.password_submit(e):
                self.update()

    def username_focus(self, e):
        self.focused = self.username

    def username_submit(self, e=None):
        if self.username.value:
            self.password.focus()
        else:
            self.username.focus()

    def password_focus(self, e):
        self.focused = self.password

    def password_submit(self, e=None):
        if self.password.value:
            return self.attempt_login(e)
        else:
            self.password.focus()
            return False

    def attempt_login(self, e=None):
        if self.show_username and not self.username.value:
            self.username.focus()
            return False
        if not self.password.value:
            self.password.focus()
            return False

        session = self.app.make_session()
        auth = self.app.get_auth_handler()
        try:
            user = auth.authenticate_user(
                session,
                self.username.value if self.show_username else None,
                self.password.value,
            )
        except:
            log.exception("user authentication error")
            session.close()
            if self.on_login_failure:
                self.on_login_failure(e)
            self.clear_login()
            return False

        user_display = str(user) if user else None
        has_perm = (
            auth.has_permission(session, user, self.perm_required) if user else False
        )
        session.close()

        if user:

            if has_perm:
                if self.on_login_success:
                    self.on_login_success(user, user_display)
                return True

            else:
                if self.on_authz_failure:
                    self.on_authz_failure(user, user_display)
                self.clear_login()

        else:
            if self.on_login_failure:
                self.on_login_failure(e)
            self.clear_login()

        return False

    def clear_login(self, e=None):
        if self.show_username:
            self.username.value = ""
        self.password.value = ""
        if self.show_username:
            self.username.focus()
        else:
            self.password.focus()
        self.update()



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Departments
"""

from wuttaweb.views import MasterView

from wuttapos.db.model import Department, Product


class DepartmentView(MasterView):
    """
    Master view for Departments
    """

    model_class = Department
    model_title = "Department"
    model_title_plural = "Departments"

    route_prefix = "departments"
    url_prefix = "/departments"

    creatable = True
    editable = True
    deletable = True

    labels = {
        "department_id": "Department ID",
    }

    grid_columns = [
        "department_id",
        "name",
        "for_products",
        "for_personnel",
        "exempt_from_gross_sales",
    ]

    form_fields = [
        "department_id",
        "name",
        "for_products",
        "for_personnel",
        "exempt_from_gross_sales",
    ]

    has_rows = True
    row_model_class = Product

    row_grid_columns = [
        "product_id",
        "brand_name",
        "description",
        "size",
        "sold_by_weight",
        "case_size",
        "special_order",
        "unit_price_reg",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")

    def get_row_grid_data(self, obj):
        department = obj
        model = self.app.model
        session = self.app.get_session(department)

        return session.query(model.Product).filter(
            model.Product.department == department
        )

    def configure_row_grid(self, grid):
        g = grid
        super().configure_row_grid(g)

        # links
        g.set_link("product_id")
        g.set_link("brand_name")
        g.set_link("description")
        g.set_link("size")

        # currency
        g.set_renderer("unit_price_reg", "currency")

        # view action
        def view_url(product, i):
            return self.request.route_url("products.view", uuid=product.uuid)

        g.add_action("view", url=view_url, icon="eye")


def defaults(config, **kwargs):
    base = globals()

    DepartmentView = kwargs.get("DepartmentView", base["DepartmentView"])
    DepartmentView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
"""
Model definition for Tenders
"""

import sqlalchemy as sa
from sqlalchemy import orm

from wuttjamaican.db import model


class Tender(model.Base):
    """
    Represents a tender (payment type) for the POS.
    """

    __tablename__ = "tender"
    __versioned__ = {}
    __wutta_hint__ = {
        "model_title": "Tender",
        "model_title_plural": "Tenders",
    }

    uuid = model.uuid_column()

    tender_id = sa.Column(
        sa.String(length=20),
        nullable=False,
        doc="""
        Unique identifier for the tender.
        """,
    )

    name = sa.Column(
        sa.String(length=50),
        nullable=False,
        doc="""
        Name for the tender type.
        """,
    )

    notes = sa.Column(
        sa.Text(),
        nullable=True,
        doc="""
        Arbitrary notes for the tender type.
        """,
    )

    is_cash = sa.Column(
        sa.Boolean(),
        nullable=False,
        doc="""
        Indicates this tender type is a form of "cash" conceptually.
        """,
    )

    is_foodstamp = sa.Column(
        sa.Boolean(),
        nullable=False,
        doc="""
        Indicates this tender type is a form of "food stamps" conceptually.
        """,
    )

    allow_cashback = sa.Column(
        sa.Boolean(),
        nullable=False,
        doc="""
        Indicates "cash back" should be allowed when overpaying with this tender.
        """,
    )

    kick_drawer = sa.Column(
        sa.Boolean(),
        nullable=False,
        doc="""
        Indicates the cash drawer should kick open when accepting this tender.
        """,
    )

    active = sa.Column(
        sa.Boolean(),
        nullable=False,
        default=True,
        doc="""
        Indicates this tender is currently active (acceptable for payment).
        """,
    )

    def __str__(self):
        return self.name or ""
